from __future__ import annotations

from typing import Dict, List

from nisystemlink.clients.core._uplink._json_model import JsonModel
from pydantic import Field

from ._file_metadata import FileMetadata
from ._link import Link


class FileQueryResponse(JsonModel):
    """The result of a file query"""

    field_links: Dict[str, Link] = Field(alias="_links")
    """The links that apply to the collection of files for a service group:
    - deleteFiles: Link to delete multiple files from the service group using a POST
    - query: Link to query for available files in the service group using a POST
    - search: Link to retrieve a filtered list of files in the service group using a GET
    - self: Link to the current service group
    - upload: Link to upload files to the service group using a POST
    """

    available_files: List[FileMetadata]
    """The list of files returned by the query"""

    total_count: int
    """The total number of files that match the query regardless of skip and take values"""



from datetime import datetime
from typing import List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel


class ScheduleTestPlanRequest(JsonModel):
    """Represents the request body content for scheduling a test plan."""

    id: str
    """Unique identifier for the test plan to be scheduled."""

    assigned_to: Optional[str] = None
    """(Optional) User or entity to whom the test plan is assigned."""

    dut_id: Optional[str] = None
    """(Optional) Identifier for the Device Under Test (DUT)."""

    system_id: Optional[str] = None
    """(Optional) Identifier for the system where the test plan will run."""

    planned_start_date_time: Optional[datetime] = None
    """(Optional) Planned start date and time for the test plan execution (ISO 8601 format)."""

    estimated_end_date_time: Optional[datetime] = None
    """(Optional) Estimated end date and time for the test plan execution (ISO 8601 format)."""

    estimated_duration_in_seconds: Optional[int] = None
    """(Optional) Estimated duration of the test plan execution in seconds."""

    fixture_ids: Optional[List[str]] = None
    """(Optional) List of fixture identifiers associated with the test plan."""



# -*- coding: utf-8 -*-

import ctypes
from ctypes import wintypes
from uuid import UUID


class GUID(ctypes.Structure):
    def __init__(self, uuid_: UUID) -> None: ...


class FOLDERID:
    AccountPictures = ...  # type: UUID
    AdminTools = ...  # type: UUID
    ApplicationShortcuts = ...  # type: UUID
    CameraRoll = ...  # type: UUID
    CDBurning = ...  # type: UUID
    CommonAdminTools = ...  # type: UUID
    CommonOEMLinks = ...  # type: UUID
    CommonPrograms = ...  # type: UUID
    CommonStartMenu = ...  # type: UUID
    CommonStartup = ...  # type: UUID
    CommonTemplates = ...  # type: UUID
    Contacts = ...  # type: UUID
    Cookies = ...  # type: UUID
    Desktop = ...  # type: UUID
    DeviceMetadataStore = ...  # type: UUID
    Documents = ...  # type: UUID
    DocumentsLibrary = ...  # type: UUID
    Downloads = ...  # type: UUID
    Favorites = ...  # type: UUID
    Fonts = ...  # type: UUID
    GameTasks = ...  # type: UUID
    History = ...  # type: UUID
    ImplicitAppShortcuts = ...  # type: UUID
    InternetCache = ...  # type: UUID
    Libraries = ...  # type: UUID
    Links = ...  # type: UUID
    LocalAppData = ...  # type: UUID
    LocalAppDataLow = ...  # type: UUID
    LocalizedResourcesDir = ...  # type: UUID
    Music = ...  # type: UUID
    MusicLibrary = ...  # type: UUID
    NetHood = ...  # type: UUID
    OriginalImages = ...  # type: UUID
    PhotoAlbums = ...  # type: UUID
    PicturesLibrary = ...  # type: UUID
    Pictures = ...  # type: UUID
    Playlists = ...  # type: UUID
    PrintHood = ...  # type: UUID
    Profile = ...  # type: UUID
    ProgramData = ...  # type: UUID
    ProgramFiles = ...  # type: UUID
    ProgramFilesX64 = ...  # type: UUID
    ProgramFilesX86 = ...  # type: UUID
    ProgramFilesCommon = ...  # type: UUID
    ProgramFilesCommonX64 = ...  # type: UUID
    ProgramFilesCommonX86 = ...  # type: UUID
    Programs = ...  # type: UUID
    Public = ...  # type: UUID
    PublicDesktop = ...  # type: UUID
    PublicDocuments = ...  # type: UUID
    PublicDownloads = ...  # type: UUID
    PublicGameTasks = ...  # type: UUID
    PublicLibraries = ...  # type: UUID
    PublicMusic = ...  # type: UUID
    PublicPictures = ...  # type: UUID
    PublicRingtones = ...  # type: UUID
    PublicUserTiles = ...  # type: UUID
    PublicVideos = ...  # type: UUID
    QuickLaunch = ...  # type: UUID
    Recent = ...  # type: UUID
    RecordedTVLibrary = ...  # type: UUID
    ResourceDir = ...  # type: UUID
    Ringtones = ...  # type: UUID
    RoamingAppData = ...  # type: UUID
    RoamedTileImages = ...  # type: UUID
    RoamingTiles = ...  # type: UUID
    SampleMusic = ...  # type: UUID
    SamplePictures = ...  # type: UUID
    SamplePlaylists = ...  # type: UUID
    SampleVideos = ...  # type: UUID
    SavedGames = ...  # type: UUID
    SavedSearches = ...  # type: UUID
    Screenshots = ...  # type: UUID
    SearchHistory = ...  # type: UUID
    SearchTemplates = ...  # type: UUID
    SendTo = ...  # type: UUID
    SidebarDefaultParts = ...  # type: UUID
    SidebarParts = ...  # type: UUID
    SkyDrive = ...  # type: UUID
    SkyDriveCameraRoll = ...  # type: UUID
    SkyDriveDocuments = ...  # type: UUID
    SkyDrivePictures = ...  # type: UUID
    StartMenu = ...  # type: UUID
    Startup = ...  # type: UUID
    System = ...  # type: UUID
    SystemX86 = ...  # type: UUID
    Templates = ...  # type: UUID
    UserPinned = ...  # type: UUID
    UserProfiles = ...  # type: UUID
    UserProgramFiles = ...  # type: UUID
    UserProgramFilesCommon = ...  # type: UUID
    Videos = ...  # type: UUID
    VideosLibrary = ...  # type: UUID
    Windows = ...  # type: UUID


def get_path(folderid: UUID, user_handle: wintypes.HANDLE = ...) -> str: ...



from typing import List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel


class DataFrame(JsonModel):
    """Data read from or to be written to a table.

    Values may be ``None`` (if the column is of type ``NULLABLE``) or encoded as
    a string in a format according to each column's datatype:

    * BOOL: One of ``"true"`` or ``"false"``, case-insensitive.
    * INT32: Any integer number in the range [-2147483648, 2147483647],
      surrounded by quotes.
    * INT64: Any integer number in the range [-9223372036854775808,
      9223372036854775807], surrounded by quotes.
    * FLOAT32: A decimal number using a period for the decimal point, optionally
      in scientific notation, in the range [-3.40282347E+38, 3.40282347E+38],
      surrounded by quotes. Not all values within the range can be represented
      with 32 bits. To preserve the exact binary encoding of the value when
      converting to a string, clients should serialize 9 digits after the
      decimal. Instead of a number, the value may be ``"NaN"`` (not a number),
      ``"Infinity"`` (positive infinity), or ``"-Infinity"`` (negative
      infinity), case-sensitive.
    * FLOAT64: A decimal number using a period for the decimal point, optionally
      in scientific notation, in the range [-1.7976931348623157E+308,
      1.7976931348623157E+308], surrounded by quotes. Not all values within the
      range can be represented with 64 bits. To preserve the exact binary
      encoding of the value when converting to a string, clients should
      serialize 17 digits after the decimal. Instead of a number, the value may
      be ``"NaN"`` (not a number), ``"Infinity"`` (positive infinity), or
      ``"-Infinity"`` (negative infinity), case-sensitive.
    * STRING: Any quoted string.
    * TIMESTAMP: A date and time with millisecond precision in ISO-8601 format
      and time zone. For example: ``"2022-08-19T16:17:30.123Z"``. If a time zone
      is not provided, UTC is assumed. If a time zone other than UTC is
      provided, the value will be converted to UTC. If more than three digits of
      fractional seconds are provided, the time will be truncated to three
      digits (i.e. milliseconds).

    The format is the same as a serialized Pandas DataFrame with orient="split"
    and index=False. See
    https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_json.html.

    When providing a DataFrame for appending rows, any of the table's columns
    not specified will receive a value of ``None``. If any such columns aren't
    nullable, an error will be returned. If the entire columns property is left
    out, each row is assumed to contain all columns in the order specified when
    the table was created.
    """

    columns: Optional[List[str]] = None
    """The names and order of the columns included in the data frame."""

    data: List[List[Optional[str]]]
    """The data for each row with the order specified in the columns property.
    Must contain a value for each column in the columns property."""



# -*- coding: utf-8 -*-

"""Implementation of SerializedTagWithAggregates."""

import datetime
from typing import Optional

from nisystemlink.clients import tag as tbase
from typing_extensions import final


@final
class SerializedTagWithAggregates:
    """Represents a generic tag value serialized to a string with optional aggregate
    values also serialized to a string.

    Clients typically do not interact with this instances of this class directly. Use a
    :class:`.TagValueReader` instead.
    """

    def __init_subclass__(cls) -> None:
        raise TypeError(
            "type 'SerializedTagWithAggregates' is not an acceptable base type"
        )

    def __init__(
        self,
        path: str,
        data_type: tbase.DataType,
        value: str,
        timestamp: Optional[datetime.datetime] = None,
        count: Optional[int] = None,
        min: Optional[str] = None,
        max: Optional[str] = None,
        mean: Optional[float] = None,
    ) -> None:
        """Initialize an instance.

        Args:
            path: The path of the tag associated with the value.
            data_type: The data type of the value serialized as a string.
            value: The value serialized as a string.
            timestamp: The timestamp associated with the value.
            count: The number of times the tag has been written, or None if the tag is
                not collecting aggregates.
            min: The minimum value of the tag serialized to a string, or None if the tag
                is not collecting aggregates or the data type of the tag does not track
                a minimum value.
            max: The maximum value of the tag serialized to a string, or None if the tag
                is not collecting aggregates or the data type of the tag does not track
                a maximum value.
            mean: The mean value of the tag, or None if the tag is not collecting
                aggregates or the data type of the tag does not track a mean value.
        """
        self._path = path
        self._data_type = data_type
        self._value = value
        self._timestamp = timestamp
        self._count = count
        self._min = min
        self._max = max
        self._mean = mean

    @property
    def data_type(self) -> tbase.DataType:  # noqa: D401
        """The data type of the value that was serialized as a string."""
        return self._data_type

    @property
    def path(self) -> str:  # noqa: D401
        """The path of the tag associated with the value."""
        return self._path

    @property
    def value(self) -> str:  # noqa: D401
        """The value of the tag serialized as a string."""
        return self._value

    @property
    def timestamp(self) -> Optional[datetime.datetime]:  # noqa: D401
        """The timestamp associated with the value, if available."""
        return self._timestamp

    @property
    def count(self) -> Optional[int]:  # noqa: D401
        """The number of times the tag has been written, or None if the tag is not collecting aggregates."""
        return self._count

    @property
    def min(self) -> Optional[str]:  # noqa: D401
        """The minimum value of the tag serialized to a string, or None if the tag is
        not collecting aggregates or the data type of the tag does not track a minimum
        value.
        """
        return self._min

    @property
    def max(self) -> Optional[str]:  # noqa: D401
        """The maximum value of the tag serialized to a string, or None if the tag is
        not collecting aggregates or the data type of the tag does not track a maximum
        value.
        """
        return self._max

    @property
    def mean(self) -> Optional[float]:  # noqa: D401
        """The mean value of the tag, or None if the tag is not collecting aggregates or
        the data type of the tag does not track a mean value.
        """
        return self._mean



# -*- coding: utf-8 -*-

"""Implementation of TagValueReader."""

from typing import Generic, Optional, TypeVar

from nisystemlink.clients import core, tag as tbase

_Any = TypeVar("_Any")


class TagValueReader(Generic[_Any]):
    """Represents the ability to read a single tag's value using an :class:`ITagReader`."""

    def __init__(self, reader: tbase.ITagReader, tag: tbase.TagData) -> None:
        """Initialize an instance.

        Args:
            reader: The :class:`ITagReader` to use when reading values.
            tag: The tag whose values will be read.
            path: The path of the tag whose values will be read.

        Raises:
            ValueError: if ``tag`` is not a tag of a valid data type and with a valid
                path.
        """
        if tag.data_type == tbase.DataType.UNKNOWN:
            raise ValueError("tag.data_type cannot be UNKNOWN")

        self._path = tag.validate_path()
        self._data_type = tag.data_type
        self.__reader = reader

    @property
    def data_type(self) -> tbase.DataType:  # noqa: D401
        """The data type of the tag associated with the value."""
        return self._data_type

    @property
    def path(self) -> str:  # noqa: D401
        """The path of the tag associated with the value."""
        return self._path

    @property
    def _reader(self) -> tbase.ITagReader:  # noqa: D401
        """The underlying :class:`ITagReader` for reading values."""
        return self.__reader

    def read(
        self, *, include_timestamp: bool = False, include_aggregates: bool = False
    ) -> Optional[tbase.TagWithAggregates[_Any]]:
        """Read the current value of the tag.

        Args:
            include_timestamp: True to include the timestamp associated with the value
                in the result.
            include_aggregates: True to include the tag's aggregate values in the result
                if the tag is set to :attr:`TagData.collect_aggregates`.

        Returns:
            The value, and the timestamp and/or aggregate values if requested, or None
            if the tag exists but doesn't have a value.

        Raises:
            ReferenceError: if the underlying reader has been closed.
            ApiException: if the API call fails.
        """
        ret = self._reader.read(
            self._path,
            include_timestamp=include_timestamp,
            include_aggregates=include_aggregates,
        )
        if ret is None:
            return None
        if ret.data_type != self.data_type:
            raise core.ApiException("Tag data type does not match")

        return ret

    async def read_async(
        self, *, include_timestamp: bool = False, include_aggregates: bool = False
    ) -> Optional[tbase.TagWithAggregates[_Any]]:
        """Read the current value of the tag.

        Args:
            include_timestamp: True to include the timestamp associated with the value
                in the result.
            include_aggregates: True to include the tag's aggregate values in the result
                if the tag is set to :attr:`TagData.collect_aggregates`.

        Returns:
            The value, and the timestamp and/or aggregate values if requested, or None
            if the tag exists but doesn't have a value.

        Raises:
            ReferenceError: if the underlying reader has been closed.
            ApiException: if the API call fails.
        """
        ret = await self._reader.read_async(
            self._path,
            include_timestamp=include_timestamp,
            include_aggregates=include_aggregates,
        )
        if ret is None:
            return None
        if ret.data_type != self.data_type:
            raise core.ApiException("Tag data type does not match")

        return ret



from typing import Dict, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel


class UpdateMetadataRequest(JsonModel):
    """The file's metadata and options for updating it."""

    replace_existing: bool
    """Determines whether the current list should be entirely replaced by the specified list
        or merged with the existing list."""

    expected_revision: Optional[int] = None
    """When specified, this is an integer that should be set to match the last known revision number of the metadata.
    If it doesn't match at the time of execution, the update request will be rejected.
    This is used to ensure that changes to this file metadata are correctly using the previous state."""

    properties: Dict[str, str]
    """The properties to set. A map of key value properties containing the metadata to be attached.

        Predefined:
            Name: This is an optional property for renaming the file. When specified, the file will be renamed."""

    workspace: Optional[str] = None
    """When specified, the workspace of the metadata will be updated to the new value."""



from enum import Enum
from typing import Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel


class StatusType(str, Enum):
    """The types of statuses that a result can have."""

    LOOPING = "LOOPING"
    SKIPPED = "SKIPPED"
    CUSTOM = "CUSTOM"
    DONE = "DONE"
    PASSED = "PASSED"
    FAILED = "FAILED"
    RUNNING = "RUNNING"
    WAITING = "WAITING"
    TERMINATED = "TERMINATED"
    ERRORED = "ERRORED"
    TIMED_OUT = "TIMED_OUT"


class Status(JsonModel):
    """Contains information about a status object."""

    status_type: StatusType
    """The type of status."""

    status_name: Optional[str] = None
    """The name of the status."""

    @staticmethod
    def LOOPING() -> "Status":
        return Status(status_type=StatusType.LOOPING)

    @staticmethod
    def SKIPPED() -> "Status":
        return Status(status_type=StatusType.SKIPPED)

    @staticmethod
    def DONE() -> "Status":
        return Status(status_type=StatusType.DONE)

    @staticmethod
    def PASSED() -> "Status":
        return Status(status_type=StatusType.PASSED)

    @staticmethod
    def FAILED() -> "Status":
        return Status(status_type=StatusType.FAILED)

    @staticmethod
    def RUNNING() -> "Status":
        return Status(status_type=StatusType.RUNNING)

    @staticmethod
    def WAITING() -> "Status":
        return Status(status_type=StatusType.WAITING)

    @staticmethod
    def TERMINATED() -> "Status":
        return Status(status_type=StatusType.TERMINATED)

    @staticmethod
    def ERRORED() -> "Status":
        return Status(status_type=StatusType.ERRORED)

    @staticmethod
    def TIMED_OUT() -> "Status":
        return Status(status_type=StatusType.TIMED_OUT)



from typing import List, Optional

from nisystemlink.clients.core import ApiError
from nisystemlink.clients.core._uplink._json_model import JsonModel
from nisystemlink.clients.testmonitor.models._result import Result
from nisystemlink.clients.testmonitor.models._update_result_request import (
    UpdateResultRequest,
)


class UpdateResultsPartialSuccess(JsonModel):
    results: List[Result]
    """The list of results that were successfully created."""

    failed: Optional[List[UpdateResultRequest]] = None
    """The list of results that were not created.
    If this is `None`, then all results were successfully created.
    """

    error: Optional[ApiError] = None
    """Error messages for results that were not created.
    If this is `None`, then all results were successfully created.
    """



from nisystemlink.clients.core._uplink._json_model import JsonModel


class CreateVirtualSystemRequest(JsonModel):
    """Model for create virtual system response containing the minion ID of the system which is created."""

    alias: str
    """Alias of the virtual system."""

    workspace: str
    """Workspace to create the virtual system in."""



from typing import List, Optional

from nisystemlink.clients.core._api_error import ApiError
from nisystemlink.clients.core._uplink._json_model import JsonModel


class DeleteAssetsResponse(JsonModel):
    """Model for delete Assets Response containing the ids of the assets which were deleted"""

    ids: Optional[List[str]] = None
    """Gets or sets array of asset identifiers which were deleted."""

    failed: Optional[List[str]] = None
    """Gets or sets array of asset identifiers that failed to delete."""

    error: Optional[ApiError] = None
    """Gets or sets error if the delete operation failed."""



"""Utilities for FileClient."""

from nisystemlink.clients.file import FileClient
from nisystemlink.clients.file.models import UpdateMetadataRequest


def rename_file(client: FileClient, file_id: str, name: str) -> None:
    """Rename a file identified by `file_id` to `name`.

    Args:
        client: The FileClient to use for the rename.
        file_id: ID of file to rename.
        name: New name for the File.
    """
    new_metadata = {"Name": name}
    rename_request = UpdateMetadataRequest(
        replace_existing=False, properties=new_metadata
    )
    client.update_metadata(metadata=rename_request, id=file_id)



from datetime import datetime
from typing import Dict, List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel

from ._dashboard import Dashboard
from ._execution_definition import ExecutionDefinition


class TestPlanTemplateBase(JsonModel):
    """Contains information about a test plan template."""

    name: Optional[str] = None
    """Name of the test plan template."""

    template_group: Optional[str] = None
    """The template group defined by the user."""

    product_families: Optional[List[str]] = None
    """Array of product families to which the test plan template belongs."""

    part_numbers: Optional[List[str]] = None
    """Array of part numbers of the products to which the test plan template belongs."""

    summary: Optional[str] = None
    """Summary of the test plan template."""

    description: Optional[str] = None
    """Description of the test plan created from this template."""

    test_program: Optional[str] = None
    """Test program name of the test plan created from this template."""

    estimated_duration_in_seconds: Optional[int] = None
    """The estimated time in seconds for executing the test plan created from this template."""

    system_filter: Optional[str] = None
    """The LINQ filter string is used to filter the potential list of
    systems capable of executing test plans created from this template.
    """

    dut_filter: Optional[str] = None
    """The LINQ filter string is used to filter the potential list of
    DUTs capable of executing test plans created from this template.
    """

    execution_actions: Optional[List[ExecutionDefinition]] = None
    """Defines the executions that will be used for test plan actions
    created from this template.
    """

    file_ids: Optional[List[str]] = None
    """Array of file IDs associated with the test plan template."""

    workspace: Optional[str] = None
    """ID of the workspace where the test plan template belongs.
    Default workspace will be taken if the value is not given.
    """

    properties: Optional[Dict[str, str]] = None
    """Properties of the test plan created from this template as key-value pairs."""

    dashboard: Optional[Dashboard] = None
    """Defines a dashboard reference for a test plan."""


class TestPlanTemplate(TestPlanTemplateBase):
    """Contains response information for test plan template."""

    __test__ = False

    id: Optional[str] = None
    """The globally unique id of the test plan template."""

    name: Optional[str] = None
    """Name of the test plan template."""

    created_by: Optional[str] = None
    """ID of the user who created the test plan template."""

    updated_by: Optional[str] = None
    """ID of the user who most recently updated the test plan template."""

    created_at: Optional[datetime] = None
    """ISO-8601 formatted timestamp indicating when the
    test plan template was created."""

    updated_at: Optional[datetime] = None
    """ISO-8601 formatted timestamp indicating when the
    test plan template was most recently updated."""



from typing import Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel


class QuerySystemsRequest(JsonModel):
    """Model for query systems request."""

    skip: Optional[int] = None
    """Gets or sets the number of systems to skip."""

    take: Optional[int] = None
    """Gets or sets number of systems to return maximum value is 1000."""

    filter: Optional[str] = None
    """The systems query filter is dynamic LINQ format.

    `id` : String representing the ID of the system.
    `createdTimestamp`: ISO-8601 formatted timestamp string specifying the
    date when the system was registered.
    `lastUpdatedTimestamp`: ISO-8601 formatted timestamp string
    specifying the last date the system was updated.
    `alias`: String representing the alias of the system.
    `activation.lastUpdatedTimestamp`: ISO-8601 formatted timestamp string
    specifying the last date the system activation was updated.
    `activation.data.activated`: Boolean representing whether the system is
    activated or not.
    `activation.data.licenseName`: String representing the name of the license.
    `activation.data.licenseVersion`: String representing the license version.
    `connected.lastUpdatedTimestamp`: ISO-8601 formatted timestamp string
    specifying the last date the system connection was updated.
    `connected.data.state`: String representing the state of the system.
    `grains.lastUpdatedTimestamp`: ISO-8601 formatted timestamp string
    specifying the last date the system grains were updated.
    `grains.data`: Dictionary of string to object representing general
    information about the system. Example: grains.data.os == "Windows"
    `packages.lastUpdatedTimestamp`: ISO-8601 formatted timestamp string
    specifying the last date the system installed packages were updated.
    `packages.data`: Dictionary representing software packages installed on the
    system.
    Example: packages.data.ni-package-manager-upgrader.version: String
    representing the installed version of ni-package-manager-upgrader package.
    `feeds.lastUpdatedTimestamp`: ISO-8601 formatted timestamp string
    specifying the last date the system configured feeds were updated.
    `feeds.data`: Dictionary representing the feeds configured on the system.
    `keywords.lastUpdatedTimestamp`: ISO-8601 formatted timestamp string
    specifying the last date the system keywords were updated.
    `keywords.data`: Array of strings representing the keywords of the system.
    Example: keywords.data.Contains("test")
    `properties.lastUpdatedTimestamp`: ISO-8601 formatted timestamp string
    specifying the last date the system properties were updated.
    `properties.data`: Dictionary of string to string representing metadata
    information about a system. Example: properties.data.owner == "admin"
    `status.data.http_connected`: Boolean representing the status of the http
    connection to the master.

    See [Dynamic Linq](https://github.com/ni/systemlink-OpenAPI-documents/wiki/Dynamic-Linq-Query-Language)
    documentation for more details.

    `"@0"`, `"@1"` etc. can be used in conjunction with the `substitutions` parameter to keep this
    query string more simple and reusable."""

    projection: Optional[str] = None
    """Gets or sets specifies the projection for resources.
    Use this field to receive specific properties of the model."""

    order_by: Optional[str] = None
    """Gets or sets the order in which data returns."""



from typing import Optional

from nisystemlink.clients.core._api_info import Operation
from nisystemlink.clients.core._uplink._json_model import JsonModel


class Operations(JsonModel):
    delete_files: Optional[Operation] = None
    download_data: Optional[Operation] = None
    list_files: Optional[Operation] = None
    query_files: Optional[Operation] = None
    update_metadata: Optional[Operation] = None
    upload_files: Optional[Operation] = None


class V1Operations(JsonModel):
    operations: Optional[Operations] = None
    """
    Available operations in the v1 version of the API:
    - deleteFiles: The ability to delete uploaded files
    - downloadData: The ability to download file data
    - listFiles: The ability to list available files and service groups
    - queryFiles: The ability to query available files and service groups
    - updateMetadata: The ability to update file metadata properties
    - uploadFiles: The ability to upload files
    """



from datetime import datetime
from typing import Dict, List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel

from ._dashboard import DashboardUrl
from ._execution_definition import ExecutionDefinition
from ._execution_event import ExecutionEvent
from ._state import State


class TestPlan(JsonModel):
    """Contains information about a test plan."""

    __test__ = False

    id: str
    """The unique identifier of the test plan."""

    template_id: Optional[str] = None
    """The identifier of the template used to create the test plan."""

    name: Optional[str] = None
    """The name of the test plan."""

    state: Optional[State] = None
    """The current state of the test plan."""

    substate: Optional[str] = None
    """The substate of the test plan, if any."""

    description: Optional[str] = None
    """The description of the test plan."""

    assigned_to: Optional[str] = None
    """The user or group assigned to the test plan."""

    work_order_id: Optional[str] = None
    """The identifier of the associated work order."""

    work_order_name: Optional[str] = None
    """The name of the associated work order."""

    part_number: Optional[str] = None
    """The part number associated with the test plan."""

    dut_id: Optional[str] = None
    """The identifier of the device under test (DUT)."""

    dut_serial_number: Optional[str] = None
    """The serial number of the device under test (DUT)."""

    test_program: Optional[str] = None
    """The test program associated with the test plan."""

    workspace: Optional[str] = None
    """The workspace to which the test plan belongs."""

    created_by: Optional[str] = None
    """The user who created the test plan."""

    updated_by: Optional[str] = None
    """The user who last updated the test plan."""

    system_id: Optional[str] = None
    """The identifier of the system used for the test plan."""

    fixture_ids: Optional[List[str]] = None
    """The list of fixture identifiers associated with the test plan."""

    planned_start_date_time: Optional[datetime] = None
    """The planned start date and time for the test plan."""

    estimated_end_date_time: Optional[datetime] = None
    """The estimated end date and time for the test plan."""

    estimated_duration_in_seconds: Optional[float] = None
    """The estimated duration of the test plan in seconds."""

    system_filter: Optional[str] = None
    """The filter used to select systems for the test plan."""

    dut_filter: Optional[str] = None
    """The filter used to select DUTs for the test plan."""

    created_at: Optional[datetime] = None
    """The date and time when the test plan was created."""

    updated_at: Optional[datetime] = None
    """The date and time when the test plan was last updated."""

    properties: Optional[Dict[str, str]] = None
    """Additional properties associated with the test plan."""

    file_ids_from_template: Optional[List[str]] = None
    """The list of file identifiers inherited from the template."""

    dashboard: Optional[DashboardUrl] = None
    """The dashboard data related to the test plan."""

    execution_actions: Optional[List[ExecutionDefinition]] = None
    """The execution actions defined for the test plan."""

    execution_history: Optional[List[ExecutionEvent]] = None
    """The execution history of the test plan."""

    dashboard_url: Optional[Dict[str, str]] = None
    """The URLs for dashboards related to the test plan."""



from typing import List, Optional

from nisystemlink.clients.core._api_error import ApiError
from nisystemlink.clients.core._uplink._json_model import JsonModel

from ._test_plan import TestPlan
from ._update_test_plan_request import UpdateTestPlanRequest


class UpdateTestPlansResponse(JsonModel):
    """Represents the response after attempting to update test plans."""

    updated_test_plans: Optional[List[TestPlan]] = None
    """List of successfully updated test plans."""

    failed_test_plans: Optional[List[UpdateTestPlanRequest]] = None
    """List of test plans that failed to update."""

    error: Optional[ApiError] = None
    """The error that occurred when updating the test plans."""



from typing import Any, Callable, Dict, List, Optional, Union

import pandas as pd
from nisystemlink.clients.spec.models._condition import (
    Condition,
    NumericConditionValue,
    StringConditionValue,
)
from nisystemlink.clients.spec.models._specification import (
    Specification,
    SpecificationLimit,
    SpecificationType,
)
from nisystemlink.clients.spec.utilities._constants import DataFrameHeaders


def summarize_conditions_as_a_string(
    conditions: List[Condition],
) -> List[Dict[str, str]]:
    """Converts the condition values to an easily readable string format that summarizes
    either of numeric or string condition.

    Args:
        conditions: List of all conditions in a spec.

    Returns:
        Conditions as a list of dictionary. The dictionary key will be
        "condition_<conditionName>(<conditionUnit>)".
        The dictionary value will be "[min: num; max: num, step: num], num, num"
        where data within the '[]' is numeric condition range and other num
        values are numeric condition discrete values.
        The dictionary value will be "str, str, str" - where str values are the
        condition discrete values for a string condition. If the condition doesn't
        have a name and value, it will be skipped.
    """
    return [
        {
            __generate_condition_column_header(condition): ", ".join(
                __serialize_condition_value(condition)
            )
            for condition in conditions
            if condition.name and condition.value
        }
    ]


def normalize_conditions_per_column(
    conditions: List[Condition],
) -> List[Dict[str, Any]]:
    """Convert conditions into list of dictionaries where dictionary key will be condition name
    and dictionary value will be condition value.

    Args:
        conditions: List of all conditions in a spec.

    Returns:
        Conditions as a list of dictionary. The key will be
        the condition name and the value will be the condition value which is
        either Numeric Condition Value, String Condition Value or None. If the condition doesn't
        have a name and value, it will be skipped.
    """
    return [
        {
            f"{DataFrameHeaders.CONDITION_COLUMN_HEADER_PREFIX}{condition.name}": condition.value
            for condition in conditions
            if condition.name and condition.value
        }
    ]


def normalize_conditions_per_row(
    conditions: List[Condition],
) -> List[Dict[str, Any]]:
    """Convert conditions into list of dictionaries where dictionary keys will be 'condition.name'
    and 'condition.value' and dictionary values will be condition name and condition value respectively.

    Args:
        conditions: List of all conditions in a spec.

    Returns:
        Conditions as a list of dictionary. The keys will be
        the 'condition.name' and 'condition.values' and the values will be the condition name and
        condition value which is either Numeric Condition Value, String Condition Value or None.
        If the condition doesn't have a name and value, it will be skipped.
        Each condition data will be logged as separate dictionary entry in the list which translates to
        separate row in the dataframe.
    """
    return [
        {"condition.name": condition.name, "condition.value": condition.value}
        for condition in conditions
        if condition.name and condition.value
    ]


def convert_specs_to_dataframe(
    specs: List[Specification],
    condition_format: Optional[
        Callable[[List[Condition]], List[Dict[str, Any]]]
    ] = normalize_conditions_per_column,
) -> pd.DataFrame:
    """Creates a Pandas DataFrame for the specs.

    Args:
        specs: List of specs.
        condition_format: A callback function which takes in a list of condition of a spec and returns
                          a list of dictionary of condition and its values. The dictionary keys
                          should be the condition name and the values should be the condition
                          value in any format you need. Dataframe rows will be constructed based on
                          these list of dictionaries. Each dictionary in the list indicates a row.
                          If there is more than one dictionary in the list, it will be considered as a new
                          row and other spec column data will be duplicated.  Keys will be used as the dataframe
                          column header and values will be used as the row cells for the
                          respective column header.
                          If not passed, condition column header will be condition name and
                          corresponding row value will be condition value.
                          For all the condition columns to be grouped together in the dataframe,
                          the dictionary key should have the prefix "condition_".
                          If condition is needed as condition per row, the public method `normalize_conditions_per_row`
                          can be provided as the callback function.
                          If condition value is needed as a string summary of condition data, the public method
                          `summarize_conditions_as_a_string` can be provided as this callback function.
                          If None is passed, conditions will not be included in the dataframe.

    Returns:
        A Pandas DataFrame with the each spec fields having a separate column.
        Following fields are split into sub-columns.
            - conditions: format of the condition columns are decided by the  `condition_format`
            argument of this function.
            - Properties: All the unique properties across all specs will be split into separate columns.
            For example, properties.property1, properties.property2, etc.
    """
    specs_dict = [
        __convert_spec_to_dict(spec=spec, condition=condition)
        for spec in specs
        for condition in (
            condition_format(spec.conditions)
            if (spec.conditions and condition_format)
            else [{}]
        )
    ]

    specs_dataframe = pd.json_normalize(specs_dict)
    specs_dataframe = __format_specs_columns(specs_dataframe=specs_dataframe)
    specs_dataframe.dropna(axis="columns", how="all", inplace=True)

    return specs_dataframe


def __convert_spec_to_dict(
    spec: Specification, condition: Dict[str, Any]
) -> Dict[str, Any]:
    """Converts a spec into dictionary.

    Args:
        spec: Spec object.
        condition: Condition as a dictionary which is added to the output spec dictionary.

    Returns:
        Spec as a dictionary with the provided condition dictionary included.
    """
    return {
        **{
            key: value
            for key, value in vars(spec).items()
            if key not in ["type", "limit", "conditions"]
        },
        **(__serialize_type(spec.type) if spec.type else {}),
        **(__serialize_limits(spec.limit) if spec.limit else {}),
        **{key: value for key, value in condition.items()},
    }


def __serialize_limits(limit: SpecificationLimit) -> Dict[str, str]:
    """Serialize limit into limit.min, limit.typical and limit.max.

    Args:
        limit: Limit of a spec.

    Returns:
        Limit as a dictionary.
    """
    return {f"limit.{key}": value for key, value in vars(limit).items()}


def __serialize_type(type: SpecificationType) -> Dict[str, str]:
    """Serialize type into it's string value.

    Args:
        type: Type of a spec.

    Returns:
        Type as a dictionary.
    """
    return {"type": type.name}


def __format_specs_columns(specs_dataframe: pd.DataFrame) -> pd.DataFrame:
    """Format specs column to group conditions and keep properties and keywords at the end.

    Args:
        specs_dataframe: Dataframe of specs.

    Returns:
        Formatted dataframe of specs.
    """
    column_headers = specs_dataframe.columns.to_list()
    standard_column_headers = [
        header for header in column_headers if __is_standard_column_header(header)
    ]
    condition_headers = [
        header for header in column_headers if __is_condition_header(header=header)
    ]
    properties_headers = [
        header for header in column_headers if __is_property_header(header=header)
    ]
    formatted_column_headers = (
        standard_column_headers
        + condition_headers
        + (["keywords"] if "keywords" in column_headers else [])
        + properties_headers
    )

    return specs_dataframe.reindex(columns=formatted_column_headers, copy=False)


def __is_standard_column_header(header: str) -> bool:
    """Check if column header is not a condition, property or keywords.

    Args:
        header: column header for specs dataframe.

    Returns:
        True if header doesn't start with condition_, properties. or keywords. Else returns false.

    """
    return not (
        __is_condition_header(header=header)
        or __is_property_header(header=header)
        or __is_keywords_header(header=header)
    )


def __is_condition_header(header: str) -> bool:
    """Check if column header is not a condition.

    Args:
        header: column header for specs dataframe.

    Returns:
        True if header contains 'condition_'. Else returns false.

    """
    return header.startswith(DataFrameHeaders.CONDITION_COLUMN_HEADER_PREFIX)


def __is_property_header(header: str) -> bool:
    """Check if column header is not a property.

    Args:
        header: column header for specs dataframe.

    Returns:
        True if header contains 'properties.'. Else returns false.

    """
    return header.startswith(DataFrameHeaders.PROPERTY_COLUMN_HEADER_PREFIX)


def __is_keywords_header(header: str) -> bool:
    """Check if column header is not a keywords.

    Args:
        header: column header for specs dataframe.

    Returns:
        True if header equals 'keywords'. Else returns false.

    """
    return header == DataFrameHeaders.KEYWORDS_COLUMN_HEADER


def __generate_condition_column_header(condition: Condition) -> str:
    """Generate column header for a condition.

    Args:
        condition: Condition object for generating column header.

    Returns:
        The column header for the given condition.
    """
    name = condition.name or ""
    unit = (
        f"({condition.value.unit})"
        if isinstance(condition.value, NumericConditionValue) and condition.value.unit
        else ""
    )

    return f"{DataFrameHeaders.CONDITION_COLUMN_HEADER_PREFIX}{name}{unit}"


def __serialize_condition_value(condition: Condition) -> List[str]:
    """Get ranges and discrete values of a condition.

    Args:
        condition: Condition for getting values.

    Returns:
        The list of values of the given condition in a specific format.
    """
    if not condition.value:
        return []

    values = []

    if isinstance(condition.value, NumericConditionValue):
        values.extend(__serialize_numeric_condition_range(value=condition.value))

    values.extend(__serialize_condition_discrete_values(value=condition.value))

    return values


def __serialize_numeric_condition_range(value: NumericConditionValue) -> List[str]:
    """Serialize ranges of a numeric condition value.

    Args:
        value: A condition's value with NumericConditionValue type.

    Returns:
        The list of ranges of the given condition where each range will be in
        string format `[min: <value>; max: <value>; step: <value>]` if the corresponding
        fields are not none.
    """
    if not value.range:
        return []

    return [
        f"""[{'; '.join(
            f'{range_key}: {range_value}'
            for range_key, range_value in vars(range).items()
            if range_value is not None
        )}]"""
        for range in value.range
    ]


def __serialize_condition_discrete_values(
    value: Union[NumericConditionValue, StringConditionValue],
) -> List[str]:
    """Serialize discrete values of a value.

    Args:
        value: A condition's value with either NumericConditionValue type or StringConditionValue type.

    Returns:
        The list of discrete values of the given value in a string format.
    """
    return [str(discrete) for discrete in (value.discrete or [])]



from nisystemlink.clients.core._uplink._json_model import JsonModel
from pydantic import Field


class Link(JsonModel):
    href: str = Field(..., examples=["/nifile/v1/service-groups"])
    """
    URI of the link
    """



from ._api_info import Operation, V1Operations
from ._condition import (
    Condition,
    ConditionRange,
    ConditionType,
    NumericConditionValue,
    StringConditionValue,
)
from ._create_specs_request import (
    CreatedSpecification,
    CreateSpecificationsPartialSuccess,
    CreateSpecificationsRequest,
    CreateSpecificationsRequestObject,
)
from ._delete_specs_request import DeleteSpecificationsPartialSuccess
from ._query_specs import (
    QuerySpecificationsRequest,
    PagedSpecifications,
    SpecificationProjection,
)
from ._specification import (
    Specification,
    SpecificationDefinition,
    SpecificationLimit,
    SpecificationType,
)
from ._update_specs_request import (
    UpdatedSpecification,
    UpdateSpecificationsPartialSuccess,
    UpdateSpecificationsRequest,
    UpdateSpecificationsRequestObject,
)

# flake8: noqa



from datetime import datetime
from typing import Any, Dict, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel


class NotebookMetadata(JsonModel):
    """Metadata for a notebook."""

    id: Optional[str] = None
    """The ID of the notebook."""

    name: Optional[str] = None
    """The name of the notebook."""

    workspace: Optional[str] = None
    """The Id of the workspace containing the notebook."""

    created_by: Optional[str] = None
    """The Id of the user that created the notebook."""

    updated_by: Optional[str] = None
    """The Id of the user that last updated the notebook."""

    created_at: Optional[datetime] = None
    """The created timestamp (ISO8601 format)."""

    updated_at: Optional[datetime] = None
    """The last updated timestamp (ISO8601 format)."""

    properties: Optional[Dict[str, str]] = None
    """A map of key value properties associated with the notebook."""

    metadata: Optional[Dict[str, Any]] = None
    """The metadata of the notebook."""

    parameters: Optional[Dict[str, Any]] = None
    """The parameters of the notebook."""



from typing import List, Optional

from nisystemlink.clients.core._uplink._with_paging import WithPaging
from nisystemlink.clients.testmonitor.models import Result


class PagedResults(WithPaging):
    """The response for a Results query containing matched results."""

    results: List[Result]
    """A list of all the results in this page."""

    total_count: Optional[int] = None
    """The total number of results that match the query."""



from __future__ import annotations

from typing import Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel

from ._platform import Platform


class CreateFeedRequest(JsonModel):
    """Create Feed Request."""

    name: str
    """The name of the feed."""
    description: Optional[str] = None
    """The description of the feed."""
    platform: Platform
    """The platform of the feed, the following package extensions are available: .nipkg for
    windows feeds, .ipk and .deb for ni-linux-rt feeds."""
    workspace: Optional[str] = None
    """The ID of the workspace this feed belongs to. If the workspace is not defined,
    the default workspace is used."""



from enum import Enum
from typing import List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel


class ExecutionSortField(str, Enum):
    """Possible fields used to sort executions."""

    NOTEBOOK_ID = "NOTEBOOK_ID"

    EXECUTION_HASH = "EXECUTION_HASH"

    TIMEOUT = "TIMEOUT"

    QUEUED_AT = "QUEUED_AT"

    STARTED_AT = "STARTED_AT"

    COMPLETED_AT = "COMPLETED_AT"

    STATUS = "STATUS"


class ExecutionField(str, Enum):
    """Possible fields in executions."""

    ID = "id"

    NOTEBOOK_ID = "notebookId"

    ORGANIZATION_ID = "orgId"

    USER_ID = "userId"

    PARAMETERS = "parameters"

    WORKSPACE_ID = "workspaceId"

    TIMEOUT = "timeout"

    STATUS = "status"

    QUEUED_AT = "queuedAt"

    STARTED_AT = "startedAt"

    COMPLETED_AT = "completedAt"

    LAST_UPDATED_AT = "lastUpdatedTimestamp"

    EXCEPTION = "exception"

    ERROR_CODE = "errorCode"

    REPORT_ID = "reportId"

    REPORT_SETTINGS = "reportSettings"

    RESULT = "result"

    SOURCE = "source"

    PRIORITY = "priority"

    RESOURCE_PROFILE = "resourceProfile"


class QueryExecutionsRequest(JsonModel):
    """Query for executions of Jupyter notebooks."""

    filter: Optional[str] = None
    """The query filter in Dynamic LINQ."""

    order_by: Optional[ExecutionSortField] = None
    """Possible fields used to sort executions."""

    descending: bool = False
    """Whether to return the executions in descending order."""

    projection: List[ExecutionField] = []
    """The projection to be applied for the items in the provider."""


class _QueryExecutionsRequest(JsonModel):
    """Query for executions of Jupyter notebooks."""

    filter: Optional[str] = None
    """The query filter in Dynamic LINQ."""

    order_by: Optional[ExecutionSortField] = None
    """Possible fields used to sort executions."""

    descending: bool = False
    """Whether to return the executions in descending order."""

    projection: Optional[str] = None
    """The projection to be applied for the items in the provider."""



from enum import Enum
from typing import List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel


class OrderBy(Enum):
    """Field by which assets can be ordered/sorted. If OrderBy is not specified, no sorting will applied."""

    LAST_UPDATED_TIMESTAMP = "LAST_UPDATED_TIMESTAMP"


class AssetField(str, Enum):
    """Model for an object describing an asset with all of its properties."""

    MODEL_NAME = "MODEL_NAME"
    MODEL_NUMBER = "MODEL_NUMBER"
    SERIAL_NUMBER = "SERIAL_NUMBER"
    VENDOR_NAME = "VENDOR_NAME"
    VENDOR_NUMBER = "VENDOR_NUMBER"
    BUS_TYPE = "BUS_TYPE"
    NAME = "NAME"
    ASSET_TYPE = "ASSET_TYPE"
    DISCOVERY_TYPE = "DISCOVERY_TYPE"
    FIRMWARE_VERSION = "FIRMWARE_VERSION"
    HARDWARE_VERSION = "HARDWARE_VERSION"
    VISA_RESOURCE_NAME = "VISA_RESOURCE_NAME"
    TEMPERATURE_SENSORS = "TEMPERATURE_SENSORS"
    SUPPORTS_SELF_CALIBRATION = "SUPPORTS_SELF_CALIBRATION"
    SUPPORTS_EXTERNAL_CALIBRATION = "SUPPORTS_EXTERNAL_CALIBRATION"
    CUSTOM_CALIBRATION_INTERVAL = "CUSTOM_CALIBRATION_INTERVAL"
    SELF_CALIBRATION = "SELF_CALIBRATION"
    IS_NI_ASSET = "IS_NI_ASSET"
    ID = "ID"
    LOCATION = "LOCATION"
    CALIBRATION_STATUS = "CALIBRATION_STATUS"
    IS_SYSTEM_CONTROLLER = "IS_SYSTEM_CONTROLLER"
    EXTERNAL_CALIBRATION = "EXTERNAL_CALIBRATION"
    WORKSPACE = "WORKSPACE"
    PROPERTIES = "PROPERTIES"
    KEYWORDS = "KEYWORDS"
    LAST_UPDATE = "LAST_UPDATE"
    FILES_IDS = "FILES_IDS"
    SUPPORTS_SELF_RESET = "SUPPORTS_SELF_RESET"
    SUPPORTS_RESET = "SUPPORTS_RESET"
    PART_NUMBER = "PART_NUMBER"
    OUT_FOR_CALIBRATION = "OUT_FOR_CALIBRATION"


class QueryAssetsRequest(JsonModel):
    """Model for object containing filters to apply when retrieving assets."""

    projection: Optional[List[AssetField]] = None
    """
    Gets or sets the projection to be used when retrieving the assets. If not specified,
    all properties will be returned.
    """

    skip: Optional[int] = None
    """Gets or sets the number of resources to skip in the result when paging."""

    take: Optional[int] = None
    """Gets or sets how many resources to return in the result, or -1 to use a default defined by the service."""

    order_by: Optional[OrderBy] = None
    """Field by which assets can be ordered/sorted. If OrderBy is not specified, no sorting will applied."""

    descending: Optional[bool] = None
    """Whether to return the assets in the descending order. If OrderBy is not specified, this property is ignored."""

    return_count: Optional[bool] = None
    """
    Gets or sets Whether to return the total number of assets which match the provided filter,
    disregarding the take value.
    """

    filter: Optional[str] = None
    """Gets or sets the filter criteria for assets. Consists of a string of queries composed using AND/OR operators."""


class _QueryAssetsRequest(JsonModel):
    """Model for object containing filters to apply when retrieving assets."""

    projection: Optional[str] = None
    """
    Gets or sets the projection to be used when retrieving the assets. If not specified,
    all properties will be returned.
    """

    skip: Optional[int] = None
    """Gets or sets the number of resources to skip in the result when paging."""

    take: Optional[int] = None
    """Gets or sets how many resources to return in the result, or -1 to use a default defined by the service."""

    order_by: Optional[OrderBy] = None
    """Field by which assets can be ordered/sorted. If OrderBy is not specified, no sorting will applied."""

    descending: Optional[bool] = None
    """Whether to return the assets in the descending order. If OrderBy is not specified, this property is ignored."""

    return_count: Optional[bool] = None
    """
    Gets or sets Whether to return the total number of assets which match the provided filter,
    disregarding the take value.
    """

    filter: Optional[str] = None
    """Gets or sets the filter criteria for assets. Consists of a string of queries composed using AND/OR operators."""



from typing import Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel

from ._data_frame import DataFrame


class AppendTableDataRequest(JsonModel):
    """Contains the rows to append and optional flags. The ``frame`` field is
    required unless ``endOfData`` is true.
    """

    frame: Optional[DataFrame] = None
    """The data frame containing the rows to append."""

    end_of_data: Optional[bool] = None
    """Whether the table should expect any additional rows to be appended in future requests."""



from datetime import datetime
from typing import Annotated, List, Literal, Optional, Union

from nisystemlink.clients.core._uplink._json_model import JsonModel
from pydantic import Field


class ExecutionEventBase(JsonModel):
    """Base class for execution events, containing common attributes such as action."""

    action: Optional[str] = None
    """The user-defined action that initiated the event."""

    triggered_at: Optional[datetime] = None
    """The time the event was triggered."""

    triggered_by: Optional[str] = None
    """The user who triggered the event."""


class NotebookExecutionEvent(ExecutionEventBase):
    """Represents an execution event that was triggered by a notebook execution."""

    type: Literal["NOTEBOOK"] = Field(default="NOTEBOOK")
    """Represents an execution event triggered by a notebook."""

    execution_id: Optional[str] = None
    """Includes the type identifier and the execution ID."""


class JobExecutionEvent(ExecutionEventBase):
    """A concrete execution event that represents an event triggered by a job."""

    type: Literal["JOB"] = Field(default="JOB")
    """Represents an execution event triggered by a job."""

    job_ids: Optional[List[str]] = None
    """Includes the type identifier and a list of job IDs."""


class ManualExecutionEvent(ExecutionEventBase):
    """A concrete execution event that represents an event triggered manually."""

    type: Literal["MANUAL"] = Field(default="MANUAL")
    """Represents an execution event triggered manually. Includes only the type identifier."""


ExecutionEvent = Annotated[
    Union[
        NotebookExecutionEvent,
        ManualExecutionEvent,
        JobExecutionEvent,
    ],
    Field(discriminator="type"),
]



"""Implementation of AssetManagementClient."""

from typing import List, Optional

from nisystemlink.clients import core
from nisystemlink.clients.core._http_configuration import HttpConfiguration
from nisystemlink.clients.core._uplink._base_client import BaseClient
from nisystemlink.clients.core._uplink._methods import post
from uplink import Field, Path, retry

from . import models


@retry(
    when=retry.when.status(408, 429, 502, 503, 504),
    stop=retry.stop.after_attempt(5),
    on_exception=retry.CONNECTION_ERROR,
)
class AssetManagementClient(BaseClient):
    def __init__(self, configuration: Optional[HttpConfiguration] = None):
        """Initialize an instance.

        Args:
            configuration: Defines the web server to connect to and information about
                how to connect. If not provided, the
                :class:`HttpConfigurationManager <nisystemlink.clients.core.HttpConfigurationManager>`
                is used to obtain the configuration.

        Raises:
            ApiException: If unable to communicate with the asset management Service.
        """
        if configuration is None:
            configuration = core.HttpConfigurationManager.get_configuration()

        super().__init__(configuration, base_path="/niapm/v1/")

    @post("assets", args=[Field("assets")])
    def create_assets(
        self, assets: List[models.CreateAssetRequest]
    ) -> models.CreateAssetsPartialSuccessResponse:
        """Create Assets.

        Args:
            assets: Array of assets that should be created.

        Returns:
            CreateAssetsPartialSuccessResponse: Array of created assets and array of failed.

        Raises:
            ApiException: If unable to communicate with the asset management service or if there are invalid
            arguments.
        """
        ...

    @post("query-assets")
    def __query_assets(
        self, query: models._QueryAssetsRequest
    ) -> models.QueryAssetsResponse:
        """Query Assets.

        Args:
            query: Object containing filters to apply when retrieving assets.

        Returns:
            QueryAssetsResponse: Assets Response containing the assets satisfying the query and
            the total count of matching assets.

        Raises:
            ApiException: If unable to communicate with the asset management service or if there are invalid
            arguments.
        """
        ...

    def query_assets(
        self, query: models.QueryAssetsRequest
    ) -> models.QueryAssetsResponse:
        """Query Assets.

        Args:
            query: Object containing filters to apply when retrieving assets.

        Returns:
            QueryAssetsResponse: Assets Response containing the assets satisfying the query and
            the total count of matching assets.

        Raises:
            ApiException: If unable to communicate with the asset management service or if there are invalid
            arguments.
        """
        projection_str = (
            f"new({', '.join(projection.name for projection in query.projection)})"
            if query.projection
            else None
        )
        query_params = {
            "filter": query.filter,
            "take": query.take,
            "skip": query.skip,
            "return_count": query.return_count,
            "order_by": query.order_by,
            "descending": query.descending,
            "projection": projection_str,
        }

        query_params = {k: v for k, v in query_params.items() if v is not None}

        query_request = models._QueryAssetsRequest(**query_params)

        return self.__query_assets(query=query_request)

    @post("delete-assets", args=[Field("ids")])
    def delete_assets(self, ids: List[str]) -> models.DeleteAssetsResponse:
        """Delete Assets.

        Args:
            ids: List of IDs of the assets to delete.

        Returns:
            DeleteAssetsResponse: Response containing the IDs of the assets that were deleted.

        Raises:
            ApiException: If unable to communicate with the asset management service or if there are invalid arguments.
        """
        ...

    @post("assets/{assetId}/file", args=[Path("assetId"), Field("fileIds")])
    def link_files(
        self, asset_id: str, file_ids: List[str]
    ) -> Optional[models.LinkFilesPartialSuccessResponse]:
        """Link files to an asset.

        Args:
            asset_id: The ID of the asset to which files should be linked.
            file_ids: The list of file IDs to link.

        Returns:
            None if all link files succeed, otherwise a response containing the IDs of files that were
            successfully linked and those that failed.

        Raises:
            ApiException: If unable to communicate with the asset management service or if there are invalid arguments.
        """
        ...



from typing import Dict, List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel


class UpdateTestPlanRequest(JsonModel):
    """Represents the content for updating a single test plan."""

    id: str
    """The unique identifier of the test plan to update."""

    name: Optional[str] = None
    """The new name for the test plan."""

    state: Optional[str] = None
    """The new state of the test plan."""

    description: Optional[str] = None
    """The new description for the test plan."""

    dut_id: Optional[str] = None
    """The device under test (DUT) identifier."""

    part_number: Optional[str] = None
    """The part number associated with the test plan."""

    assigned_to: Optional[str] = None
    """The user or group assigned to the test plan."""

    test_program: Optional[str] = None
    """The test program associated with the test plan."""

    properties: Optional[Dict[str, str]] = None
    """Additional properties for the test plan as key-value pairs."""

    workspace: Optional[str] = None
    """The workspace to which the test plan belongs."""

    work_order_id: Optional[str] = None
    """The work order identifier associated with the test plan."""

    file_ids_from_template: Optional[List[str]] = None
    """List of file IDs from the template to associate with the test plan."""



from typing import List, Optional

from nisystemlink.clients.core._uplink._with_paging import WithPaging
from nisystemlink.clients.testmonitor.models._step import Step


class PagedSteps(WithPaging):
    """The response for a Steps query containing matched steps."""

    steps: List[Step]
    """A list of all the steps in this page."""

    total_count: Optional[int] = None
    """The total number of steps that match the query."""



from nisystemlink.clients.systems.models._create_virtual_systems_request import (
    CreateVirtualSystemRequest,
)
from nisystemlink.clients.systems.models._create_virtual_systems_response import (
    CreateVirtualSystemResponse,
)
from nisystemlink.clients.systems.models._query_systems_request import (
    QuerySystemsRequest,
)
from nisystemlink.clients.systems.models._query_systems_response import (
    QuerySystemsResponse,
)
from nisystemlink.clients.systems.models._remove_systems_response import (
    RemoveSystemsResponse,
)

# flake8: noqa



from typing import List, Optional

from nisystemlink.clients.core import ApiError
from nisystemlink.clients.core._uplink._json_model import JsonModel
from nisystemlink.clients.product.models._product import Product
from nisystemlink.clients.product.models._product_request import CreateProductRequest


class CreateProductsPartialSuccess(JsonModel):
    products: List[Product]
    """The list of products that were successfully created."""

    failed: Optional[List[CreateProductRequest]] = None
    """The list of products that were not created.

    If this is `None`, then all products were successfully created.
    """

    error: Optional[ApiError] = None
    """Error messages for products that were not created.

    If this is `None`, then all products were successfully created.
    """



from typing import Any, Iterator


class IteratorFileLike:
    """A file-like object adapter that wraps a python iterator, providing a way to
    read from the iterator as if it was a file.
    """

    def __init__(self, iterator: Iterator[Any]):
        self._iterator = iterator
        self._buffer = b""

    def read(self, size: int = -1) -> bytes:
        """Read at most `size` bytes from the file-like object. If `size` is not
        specified or is negative, read until the iterator is exhausted and
        returns all bytes or characters read.
        """
        while size < 0 or len(self._buffer) < size:
            try:
                chunk = next(self._iterator)
                self._buffer += chunk
            except StopIteration:
                break
        if size < 0:
            data = self._buffer
            self._buffer = b""
        else:
            data = self._buffer[:size]
            self._buffer = self._buffer[size:]
        return data



"""Implementation of Product Client"""

from typing import List, Optional

from nisystemlink.clients import core
from nisystemlink.clients.core._uplink._base_client import BaseClient
from nisystemlink.clients.core._uplink._methods import delete, get, post
from uplink import Field, Query, retry, returns

from . import models


@retry(
    when=retry.when.status(408, 429, 502, 503, 504),
    stop=retry.stop.after_attempt(5),
    on_exception=retry.CONNECTION_ERROR,
)
class ProductClient(BaseClient):
    def __init__(self, configuration: Optional[core.HttpConfiguration] = None):
        """Initialize an instance.

        Args:
            configuration: Defines the web server to connect to and information about
                how to connect. If not provided, the
                :class:`HttpConfigurationManager <nisystemlink.clients.core.HttpConfigurationManager>`
                is used to obtain the configuration.

        Raises:
            ApiException: if unable to communicate with the Product Service.
        """
        if configuration is None:
            configuration = core.HttpConfigurationManager.get_configuration()
        super().__init__(configuration, base_path="/nitestmonitor/v2/")

    @post("products", args=[Field("products")])
    def create_products(
        self, products: List[models.CreateProductRequest]
    ) -> models.CreateProductsPartialSuccess:
        """Creates one or more products and returns errors for failed creations.

        Args:
            products: A list of products to attempt to create.

        Returns: A list of created products, products that failed to create, and errors for
            failures.

        Raises:
            ApiException: if unable to communicate with the ``/nitestmonitor`` service of provided invalid
                arguments.
        """
        ...

    @get(
        "products",
        args=[Query("continuationToken"), Query("take"), Query("returnCount")],
    )
    def get_products_paged(
        self,
        continuation_token: Optional[str] = None,
        take: Optional[int] = None,
        return_count: Optional[bool] = None,
    ) -> models.PagedProducts:
        """Reads a list of products.

        Args:
            continuation_token: The token used to paginate results.
            take: The number of products to get in this request.
            return_count: Whether or not to return the total number of products available.

        Returns:
            A list of products.

        Raises:
            ApiException: if unable to communicate with the ``/nitestmonitor`` Service
                or provided an invalid argument.
        """
        ...

    @get("products/{id}")
    def get_product(self, id: str) -> models.Product:
        """Retrieves a single product by id.

        Args:
            id (str): Unique ID of a products.

        Returns:
            The single product matching `id`

        Raises:
            ApiException: if unable to communicate with the ``/nitestmonitor`` Service
                or provided an invalid argument.
        """
        ...

    @post("query-products")
    def query_products_paged(
        self, query: models.QueryProductsRequest
    ) -> models.PagedProducts:
        """Queries for products that match the filter.

        Args:
            query : The query contains a DynamicLINQ query string in addition to other details
                about how to filter and return the list of products.

        Returns:
            A paged list of products with a continuation token to get the next page.

        Raises:
            ApiException: if unable to communicate with the ``/nitestmonitor`` Service or provided invalid
                arguments.
        """
        ...

    @returns.json  # type: ignore
    @post("query-product-values")
    def query_product_values(
        self, query: models.QueryProductValuesRequest
    ) -> List[str]:
        """Queries for products that match the query and returns a list of the requested field.

        Args:
            query : The query for the fields you want.

        Returns:
            A list of the values of the field you requested.

        Raises:
            ApiException: if unable to communicate with the ``/nitestmonitor`` Service or provided
            invalid arguments.
        """
        ...

    @post("update-products", args=[Field("products"), Field("replace")])
    def update_products(
        self, products: List[models.UpdateProductRequest], replace: bool = False
    ) -> models.CreateProductsPartialSuccess:
        """Updates a list of products with optional field replacement.

        Args:
            `products`: A list of products to update. Products are matched for update by id.
            `replace`: Replace the existing fields instead of merging them. Defaults to `False`.
                If this is `True`, then `keywords` and `properties` for the product will be
                    replaced by what is in the `products` provided in this request.
                If this is `False`, then the `keywords` and `properties` in this request will
                    merge with what is already present in the server resource.

        Returns: A list of updates products, products that failed to update, and errors for
            failures.

        Raises:
            ApiException: if unable to communicate with the ``/nitestmonitor`` Service
                or provided an invalid argument.
        """
        ...

    @delete("products/{id}")
    def delete_product(self, id: str) -> None:
        """Deletes a single product by id.

        Args:
            id (str): Unique ID of a product.

        Raises:
            ApiException: if unable to communicate with the ``/nitestmonitor`` Service
                or provided an invalid argument.
        """
        ...

    @post("delete-products", args=[Field("ids")])
    def delete_products(
        self, ids: List[str]
    ) -> Optional[models.DeleteProductsPartialSuccess]:
        """Deletes multiple products.

        Args:
            ids (List[str]): List of unique IDs of products.

        Returns:
            A partial success if any products failed to delete, or None if all
            products were deleted successfully.

        Raises:
            ApiException: if unable to communicate with the ``/nitestmonitor`` Service
                or provided an invalid argument.
        """
        ...



from __future__ import annotations

from typing import List

from nisystemlink.clients.core._uplink._json_model import JsonModel

from ._feed import Feed


class QueryFeedsResponse(JsonModel):
    """Query Feeds response."""

    feeds: List[Feed]
    """A collection of feeds"""



from enum import Enum


class State(Enum):
    """The state of the test plan."""

    NEW = "NEW"
    DEFINED = "DEFINED"
    REVIEWED = "REVIEWED"
    SCHEDULED = "SCHEDULED"
    IN_PROGRESS = "IN_PROGRESS"
    PENDING_APPROVAL = "PENDING_APPROVAL"
    CLOSED = "CLOSED"
    CANCELED = "CANCELED"



from typing import List, Optional

from nisystemlink.clients import core
from nisystemlink.clients.core._http_configuration import HttpConfiguration
from nisystemlink.clients.core._uplink._base_client import BaseClient
from nisystemlink.clients.core._uplink._methods import get, post
from nisystemlink.clients.test_plan import models
from uplink import Field, retry


@retry(
    when=retry.when.status(408, 429, 502, 503, 504),
    stop=retry.stop.after_attempt(5),
    on_exception=retry.CONNECTION_ERROR,
)
class TestPlanClient(BaseClient):
    __test__ = False

    def __init__(self, configuration: Optional[HttpConfiguration] = None):
        """Initialize an instance.

        Args:
            configuration: Defines the web server to connect to and information about
                how to connect. If not provided, the
                :class:`HttpConfigurationManager <nisystemlink.clients.core.HttpConfigurationManager>`
                is used to obtain the configuration.

        Raises:
            ApiException: if unable to communicate with the WorkOrder Service.
        """
        if configuration is None:
            configuration = core.HttpConfigurationManager.get_configuration()

        super().__init__(configuration, base_path="/niworkorder/v1/")

    @get("testplans/{test_plan_id}")
    def get_test_plan(self, test_plan_id: str) -> models.TestPlan:
        """Retrieve a test plan by its ID.

        Args:
            test_plan_id: The ID of the test plan to retrieve.

        Returns:
            The TestPlan object corresponding to the given ID.
        """
        ...

    @post("testplans", args=[Field("testPlans")])
    def create_test_plans(
        self, test_plans: List[models.CreateTestPlanRequest]
    ) -> models.CreateTestPlansPartialSuccessResponse:
        """Create a new test plan.

        Args:
            test_plan: The test plans to create.

        Returns:
            The created test plan object.
        """
        ...

    @post("delete-testplans", args=[Field("ids")])
    def delete_test_plans(self, ids: List[str]) -> None:
        """Delete test plans by IDs.

        Args:
            test_plan_ids: A list of test plan IDs to delete.

        Returns:
            None
        """
        ...

    @post("query-testplans")
    def query_test_plans(
        self, query_request: models.QueryTestPlansRequest
    ) -> models.PagedTestPlans:
        """Query test plans.

        Args:
            query: The query to execute.

        Returns:
            A PagedTestPlans object containing test plans that match the query.
        """
        ...

    @post("schedule-testplans")
    def schedule_test_plans(
        self, schedule_request: models.ScheduleTestPlansRequest
    ) -> models.ScheduleTestPlansResponse:
        """Schedule a test plan.

        Args:
            schedule: The schedule to apply to the test plan.

        Returns:
            A ScheduleTestPlansResponse object containing the scheduled test plan.
        """
        ...

    @post("update-testplans")
    def update_test_plans(
        self, update_request: models.UpdateTestPlansRequest
    ) -> models.UpdateTestPlansResponse:
        """Update a test plan.

        Args:
            test_plan: The test plan to update.

        Returns:
            The updated test plan object.
        """
        ...

    @post("testplan-templates", args=[Field("testPlanTemplates")])
    def create_test_plan_templates(
        self, test_plan_templates: List[models.CreateTestPlanTemplateRequest]
    ) -> models.CreateTestPlanTemplatePartialSuccessResponse:
        """Creates one or more test plan template and return errors for failed creations.

        Args:
            test_plan_templates: A list of test plan templates to attempt to create.

        Returns: A list of created test plan templates, test plan templates that failed to create, and errors for
                 failures.

        Raises: ApiException: if unable to communicate with the `/niworkorder` service of provided invalid
                arguments.
        """
        ...

    @post("query-testplan-templates")
    def query_test_plan_templates(
        self, query_test_plan_templates: models.QueryTestPlanTemplatesRequest
    ) -> models.PagedTestPlanTemplates:
        """Queries one or more test plan templates and return errors for failed queries.

        Returns: A list of test plan templates, based on the query and errors for the wrong query.

        Raises: ApiException: if unable to communicate with the `/niworkorder` service of provided invalid
                arguments.
        """
        ...

    @post("delete-testplan-templates", args=[Field("ids")])
    def delete_test_plan_templates(
        self, ids: List[str]
    ) -> Optional[models.DeleteTestPlanTemplatesPartialSuccessResponse]:
        """Deletes one or more test plan templates and return errors for failed deletion.

        Returns:
            A partial success if any test plan templates failed to delete, or None if all
            test plan templates were deleted successfully.

        Raises: ApiException: if unable to communicate with the `/niworkorder` service of provided invalid
            arguments.
        """
        ...



from typing import List, Optional

from nisystemlink.clients.core._api_error import ApiError
from nisystemlink.clients.core._uplink._json_model import JsonModel

from ._create_test_plan_template_request import CreateTestPlanTemplateRequest
from ._test_plan_templates import TestPlanTemplate


class CreateTestPlanTemplatePartialSuccessResponse(JsonModel):

    created_test_plan_templates: Optional[List[TestPlanTemplate]] = None
    """The list of test plan templates that were successfully created."""

    failed_test_plan_templates: Optional[List[CreateTestPlanTemplateRequest]] = None
    """The list of test plan templates that were not created.

    If this is `None`, then all test plan templates were successfully created.
    """

    error: Optional[ApiError] = None
    """Error messages for test plan templates that were not created.

    If this is `None`, then all test plan templates were successfully created.
    """



from typing import List, Optional

from nisystemlink.clients.core._uplink._with_paging import WithPaging

from ._notebook_metadata import NotebookMetadata


class PagedNotebooks(WithPaging):
    """Model for a query notebooks response."""

    notebooks: Optional[List[NotebookMetadata]] = None
    """The list of notebooks."""



# -*- coding: utf-8 -*-

"""Implementation of HttpTagSubscription."""

import datetime
import weakref
from typing import Iterable, List, Optional

from nisystemlink.clients import core, tag as tbase
from nisystemlink.clients.core._internal._http_client import HttpClient
from nisystemlink.clients.core._internal._timestamp_utilities import TimestampUtilities
from nisystemlink.clients.tag._core._manual_reset_timer import ManualResetTimer
from nisystemlink.clients.tag._core._serialized_tag_with_aggregates import (
    SerializedTagWithAggregates,
)
from nisystemlink.clients.tag._core._serialized_tag_with_aggregates_reader import (
    SerializedTagWithAggregatesReader,
)
from typing_extensions import final


@final
class HttpTagSubscription(tbase.TagSubscription):
    _DEFAULT_POLLING_INTERVAL_MILLISECONDS = 5000.0

    __MAGIC = object()

    def __init_subclass__(cls) -> None:
        raise TypeError("type 'HttpTagSubscription' is not an acceptable base type")

    @classmethod
    def create(
        cls,
        client: HttpClient,
        paths: Iterable[str],
        update_timer: Optional[ManualResetTimer] = None,
        heartbeat_timer: Optional[ManualResetTimer] = None,
    ) -> "HttpTagSubscription":
        """Create an :class:`HttpTagSubscription` with a custom heartbeat timer for testing purposes.

        Args:
            client: The HTTP client object for communicating with the server.
            paths: The tag path queries to include in the subscription.
            update_timer: A timer for polling for updates on the server, or None to use
                a default timer.
            heartbeat_timer: A timer for sending a heartbeat to keep the subscription
                alive, or None to use a default timer.

        Returns:
            The created subscription.

        Raises:
            ValueError: if ``paths`` is None.
            ApiException: if the API call fails.
        """
        subscription = HttpTagSubscription(
            cls.__MAGIC, client, paths, update_timer, heartbeat_timer
        )
        subscription._initialize()
        return subscription

    @classmethod
    async def create_async(
        cls,
        client: HttpClient,
        paths: Iterable[str],
        update_timer: Optional[ManualResetTimer] = None,
        heartbeat_timer: Optional[ManualResetTimer] = None,
    ) -> "HttpTagSubscription":
        """Asynchronously create an :class:`HttpTagSubscription` with a custom heartbeat timer for testing purposes.

        Args:
            client: The HTTP client object for communicating with the server.
            paths: The tag path queries to include in the subscription.
            update_timer: A timer for polling for updates on the server, or None to use
                a default timer.
            heartbeat_timer: A timer for sending a heartbeat to keep the subscription
                alive, or None to use a default timer.

        Returns:
            A task representing the asynchronous operation. On completion, contains the
            created subscription.

        Raises:
            ValueError: if ``paths`` is None.
            ApiException: if the API call fails.
        """
        subscription = HttpTagSubscription(
            cls.__MAGIC, client, paths, update_timer, heartbeat_timer
        )
        await subscription._initialize_async()
        return subscription

    def __init__(
        self,
        magic: object,
        client: HttpClient,
        paths: Iterable[str],
        update_timer: Optional[ManualResetTimer] = None,
        heartbeat_timer: Optional[ManualResetTimer] = None,
    ) -> None:
        assert (
            magic is self.__MAGIC
        ), "Do not construct an HttpTagSubscription directly. Use create() instead."
        super().__init__(paths, heartbeat_timer)
        self._api = client.at_uri("/nitag/v2/subscriptions")
        if update_timer is not None:
            self._update_timer = update_timer
        else:
            self._update_timer = ManualResetTimer(
                datetime.timedelta(
                    milliseconds=self._DEFAULT_POLLING_INTERVAL_MILLISECONDS
                )
            )
            # _exit_stack is instantiated in the base class
            self._exit_stack.enter_context(self._update_timer)

        callback_ref = weakref.WeakMethod(self._update_timer_elapsed)  # type: ignore

        def callback() -> None:
            actual_callback = callback_ref()  # type: ignore
            if actual_callback:
                actual_callback()

        self._update_timer_handler = callback
        self._update_timer.elapsed += self._update_timer_handler
        self._token = None  # type: Optional[str]

    # Base class implementation is sufficient:
    #   def __enter__(self):
    #   async def __aenter__(self):
    #   def __exit__(self, exc_type, exc, traceback):
    #   async def __aexit__(self, exc_type, exc, traceback):

    def _close_internal(self) -> None:
        if self._token is None:
            return

        try:
            self._update_timer.stop()
            self._update_timer.elapsed -= self._update_timer_handler
            self._api.delete("/{id}", params={"id": self._token})
            self._token = None
        except core.ApiException:
            pass

    async def _close_internal_async(self) -> None:
        if self._token is None:
            return

        try:
            self._update_timer.stop()
            self._update_timer.elapsed -= self._update_timer_handler
            await self._api.as_async.delete("/{id}", params={"id": self._token})
            self._token = None
        except core.ApiException:
            pass

    def _create_subscription_on_server(self, paths: List[str]) -> None:
        response, http_response = self._api.post(
            "", data={"updatesOnly": True, "tags": paths}
        )

        if response is None or response.get("subscriptionId") is None:
            raise tbase.TagManager.invalid_response(http_response)

        token = response["subscriptionId"]

        # We only want to expose real changes to tags, but the subscription service
        # initially sends metadata and current values for every tag in the subscription.
        # For our API, we want clients to query using a selection prior to creating the
        # subscription if they want current information.
        self._api.get("/{id}/values/current", params={"id": token})
        self._token = token
        self._update_timer.start()

    async def _create_subscription_on_server_async(self, paths: List[str]) -> None:
        response, http_response = await self._api.as_async.post(
            "", data={"updatesOnly": True, "tags": paths}
        )

        if response is None or response.get("subscriptionId") is None:
            raise tbase.TagManager.invalid_response(http_response)

        token = response["subscriptionId"]

        # We only want to expose real changes to tags, but the subscription service
        # initially sends metadata and current values for every tag in the subscription.
        # For our API, we want clients to query using a selection prior to creating the
        # subscription if they want current information.
        await self._api.as_async.get("/{id}/values/current", params={"id": token})
        self._token = token
        self._update_timer.start()

    def _send_heartbeat(self) -> None:
        assert self._token is not None
        self._api.put("/{id}/heartbeat", params={"id": self._token})

    def _update_timer_elapsed(self) -> None:
        try:
            token = self._token
            if token is None:
                return

            try:
                response, _ = self._api.get(
                    "/{id}/values/current", params={"id": token}
                )
            except core.ApiException:
                return

            if response is None:
                return

            subscriptions = response.get("subscriptionUpdates")
            if subscriptions is None:
                return

            for subscription in subscriptions:
                if subscription is None:
                    continue

                updates = subscription.get("updates")
                if updates is None:
                    continue

                for update in updates:
                    if update is None:
                        continue

                    tag = update.get("tag")
                    timestamp = update.get("timestamp")
                    if tag is None or timestamp is None:
                        continue

                    tag = tbase.TagData.from_json_dict(tag)
                    try:
                        tag.validate_path()
                    except ValueError:
                        continue
                    aggregates = update.get("aggregates") or {}
                    if tag.data_type == tbase.DataType.UNKNOWN:
                        self._on_tag_changed(tag, None)
                    else:
                        value = SerializedTagWithAggregates(
                            tag.path,
                            tag.data_type,
                            update.get("value"),
                            TimestampUtilities.str_to_datetime(timestamp),
                            aggregates.get("count"),
                            aggregates.get("min"),
                            aggregates.get("max"),
                            (
                                float(aggregates["avg"])
                                if aggregates.get("avg") is not None
                                else None
                            ),
                        )
                        reader = tbase.TagValueReader(
                            SerializedTagWithAggregatesReader(value), tag
                        )  # type: tbase.TagValueReader
                        self._on_tag_changed(tag, reader)
        finally:
            self._update_timer.start()



from typing import Dict, List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel
from pydantic import ConfigDict, Field


class Measurement(JsonModel):
    name: Optional[str] = None
    status: Optional[str] = None
    measurement: Optional[str] = None
    lowLimit: Optional[str] = None
    highLimit: Optional[str] = None
    units: Optional[str] = None
    comparisonType: Optional[str] = None
    model_config = ConfigDict(extra="allow")
    __pydantic_extra__: Dict[str, str] = Field(init=False)


class StepData(JsonModel):
    text: Optional[str] = None
    """Text string describing the output data."""

    parameters: Optional[List[Measurement]] = None
    """List of properties objects."""



from typing import List, Optional

from nisystemlink.clients.core._api_error import ApiError
from nisystemlink.clients.core._uplink._json_model import JsonModel

from ._execution import Execution


class CreatedExecution(Execution):
    cached_result: bool
    """Returns true if the execution is returned from cache"""


class CreateExecutionsResponse(JsonModel):
    """Model for response to a request to create an execution."""

    error: Optional[ApiError] = None
    """The error that occurred during the request, if any."""

    executions: Optional[List[CreatedExecution]] = None
    """Gets or sets the collection of authorized executions."""



from ._file_metadata import FileMetadata
from ._file_query_order_by import FileQueryOrderBy
from ._file_query_response import FileQueryResponse
from ._link import Link
from ._operations import V1Operations
from ._update_metadata import UpdateMetadataRequest

# flake8: noqa



from typing import Any

from nisystemlink.clients.core._uplink._json_model import JsonModel


class NamedValue(JsonModel):
    name: str
    """The name of the value."""

    value: Any = None
    """The value."""



from datetime import datetime
from typing import Dict, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel
from pydantic import Field

from ._link import Link


class FileMetadata(JsonModel):
    field_links: Optional[Dict[str, Link]] = Field(None, alias="_links")
    """
    The links to access and manipulate the file:
    - data: Link to download the file using a GET request
    - delete: Link to delete the file using a DELETE request
    - self: Link to the file's metadata
    - updateMetadata: Link to update the file's metadata using a POST request
    """

    created: Optional[datetime] = None
    """
    The date and time the file was created in the file service.

    example :2018-05-15T18:54:27.519Z
    """

    id: Optional[str] = None
    """
    The file's ID within the service group.

    example: "5afb2ce3741fe11d88838cc9"
    """

    path: Optional[str] = None
    """
    The path to the file on the server.  This field is returned only if
    the user has associated privileges to view file paths.

    example: C:\temp\4afb2ce3741fe11d88838cc9.txt
    """

    properties: Optional[Dict[str, str]] = None
    """
    The file's properties
    Example - {"Name": "myfile.txt", "MyProperty": "Value"}
    """

    meta_data_revision: Optional[int] = None
    """
    The file's properties revision number. When a file is uploaded, the revision number starts at 1.
    Every time metadata is updated, the revision number is incremented by 1.
    """

    service_group: Optional[str] = None
    """
    The service group that owns the file
    """

    size: Optional[int] = None
    """
    The 32-bit file size in bytes. If the value is larger than a 32-bit integer,
    this value is -1 and the size64 parameter contains the correct value.
    """

    size64: Optional[int] = None
    """
    The 64-bit file size in bytes
    """

    workspace: Optional[str] = None
    """
    The workspace the file belongs to
    """

    last_updated_timestamp: Optional[datetime] = None
    """
    The date and time the file was last updated in the file service.

    example: 2018-05-15T18:54:27.519Z
    """



from datetime import datetime
from typing import Dict, List

from nisystemlink.clients.core._uplink._json_model import JsonModel

from ._column import Column


class TableMetadata(JsonModel):
    """Contains information about a table, including its properties and column definitions."""

    columns: List[Column]
    """The list of columns in the table."""

    created_at: datetime
    """The date and time the table was created."""

    id: str
    """The table's unique identifier."""

    metadata_modified_at: datetime
    """The date and time the table's metadata was last modified."""

    metadata_revision: int
    """The table's metadata revision number, incremented each time the metadata is modified."""

    name: str
    """The name associated with the table."""

    properties: Dict[str, str]
    """User-defined properties associated with the table."""

    row_count: int
    """The number of rows in the table."""

    rows_modified_at: datetime
    """The date and time the table's data was last modified."""

    supports_append: bool
    """Whether the table supports appending additional rows of data."""

    workspace: str
    """The workspace the table belongs to."""



from ._test_plan_templates import TestPlanTemplateBase


class CreateTestPlanTemplateRequest(TestPlanTemplateBase):
    """Contains information about a test plan template request."""

    name: str
    """Name of the test plan template."""

    template_group: str
    """The template group defined by the user."""



from ._query_assets_response import QueryAssetsResponse
from ._asset import Asset
from ._create_asset_request import CreateAssetRequest
from ._create_assets_partial_success_response import CreateAssetsPartialSuccessResponse
from ._delete_assets_response import DeleteAssetsResponse
from ._query_assets_request import AssetField, QueryAssetsRequest, _QueryAssetsRequest
from ._query_assets_response import QueryAssetsResponse
from ._asset_location import (
    AssetLocation,
    AssetLocationForCreate,
    AssetPresence,
    AssetPresenceWithSystemConnection,
    AssetPresenceStatus,
)
from ._asset_calibration import (
    CalibrationMode,
    CalibrationStatus,
    SelfCalibration,
    TemperatureSensor,
    ExternalCalibration,
)
from ._asset_types import AssetBusType, AssetDiscoveryType, AssetType
from ._link_files_partial_success_response import LinkFilesPartialSuccessResponse

# flake8: noqa



from typing import List, Optional

from nisystemlink.clients.testmonitor.models._create_steps_request import (
    BaseStepRequest,
)


class UpdateStepRequest(BaseStepRequest):
    name: Optional[str] = None
    """Step name."""

    children: Optional[List["UpdateStepRequest"]] = None
    """Nested child steps."""



from datetime import datetime
from typing import Dict, List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel
from nisystemlink.clients.testmonitor.models._status import Status


class UpdateResultRequest(JsonModel):
    """Contains information about a result."""

    status: Optional[Status] = None
    """The status of the result."""

    started_at: Optional[datetime] = None
    """The time that the result started."""

    program_name: Optional[str] = None
    """The name of the program that generated this result."""

    id: str
    """The globally unique id of the result."""

    system_id: Optional[str] = None
    """The id of the system that generated this result."""

    host_name: Optional[str] = None
    """The name of the host that generated this result."""

    part_number: Optional[str] = None
    """The part number is the unique identifier of a product within a single org."""

    serial_number: Optional[str] = None
    """The serial number of the system that generated this result."""

    total_time_in_seconds: Optional[float] = None
    """The total time that the result took to run in seconds."""

    keywords: Optional[List[str]] = None
    """A list of keywords that categorize this result."""

    properties: Optional[Dict[str, Optional[str]]] = None
    """A list of custom properties for this result."""

    operator: Optional[str] = None
    """The operator that ran the result."""

    file_ids: Optional[List[str]] = None
    """A list of file ids that are attached to this result."""

    data_table_ids: Optional[List[str]] = None
    """A list of data table ids that are attached to this result."""

    workspace: Optional[str] = None
    """The id of the workspace that this product belongs to."""



from typing import Optional

from nisystemlink.clients.core._uplink._with_paging import WithPaging


class QueryNotebookRequest(WithPaging):
    """Model for a query notebooks request."""

    filter: Optional[str] = None
    """
    The filter criteria for notebook, consisting of a string of queries composed using AND/OR operators.
    String values need to be enclosed in double quotes. Parenthesis can be used within the filter
    to better define the order of operations.
    Filter syntax: '[property name][operator][operand] and [property name][operator][operand]' Operators:

      Equals operator '='. Example: 'x = y'
      Not equal operator '!='. Example: 'x != y'
      Greater than operator '>'. Example: 'x > y'
      Greater than or equal operator '>='. Example: 'x >= y'
      Less than operator '<'. Example: 'x < y'
      Less than or equal operator '<='. Example: 'x <= y'
      Logical AND operator 'and'. Example: 'x and y'
      Logical OR operator 'or'. Example: 'x or y'
      Starts with operator '.StartsWith()', used to check whether a string starts with another string.
      Example: 'x.StartsWith(y)'
      Does not start with operator '!.StartsWith()', used to check whether a string does not start with another string.
      Example: '!x.StartsWith(y)'
      String null or empty 'string.IsNullOrEmpty()', used to check whether a string is null or empty.
      Example: 'string.IsNullOrEmpty(x)'
      String is not null or empty '!string.IsNullOrEmpty()', used to check whether a string is not null or empty.
      Example: '!string.IsNullOrEmpty(x)'
        Valid notebook properties that can be used in the filter:
      id
      name
      properties.interface
      workspace
    """

    take: Optional[int] = None
    """The maximum number of notebooks to return."""



class DataFrameHeaders:
    STATUS_TYPE_SUMMARY_HEADER_PREFIX = "status_type_summary."

    PROPERTY_COLUMN_HEADER_PREFIX = "properties."

    CATEGORY_COLUMN_HEADERS = [
        "general",
        "inputs",
        "outputs",
        "data",
        "properties",
    ]



from typing import List

import pandas as pd
from nisystemlink.clients.product.models import Product
from pandas import DataFrame


def convert_products_to_dataframe(products: List[Product]) -> DataFrame:
    """Converts a list of products into a normalized dataframe.

    Args:
        products (List[Product]): A list of products

    Returns:
        DataFrame:
            - A Pandas DataFrame containing the product data. The DataFrame would consist of all the
            fields in the input products.
            - A new column would be created for unique properties across all products. The property
            columns would be named in the format `properties.property_name`.
    """
    products_dict_representation = [
        product.model_dump(exclude_none=True) for product in products
    ]
    normalized_products_dataframe = pd.json_normalize(
        products_dict_representation, sep="."
    )

    return normalized_products_dataframe



from datetime import datetime
from typing import Dict, List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel
from nisystemlink.clients.testmonitor.models._named_value import NamedValue
from nisystemlink.clients.testmonitor.models._step import Status
from nisystemlink.clients.testmonitor.models._step_data import StepData


class BaseStepRequest(JsonModel):
    step_id: Optional[str]
    """Step ID."""

    result_id: str
    """Result ID."""

    parent_id: Optional[str] = None
    """Parent step ID."""

    data: Optional[StepData] = None
    """Data returned by the test step."""

    data_model: Optional[str] = None
    """Data model for the step."""

    started_at: Optional[datetime] = None
    """ISO-8601 formatted timestamp indicating when the test result began."""

    status: Optional[Status] = None
    """The status of the step."""

    step_type: Optional[str] = None
    """Step type."""

    total_time_in_seconds: Optional[float] = None
    """Total number of seconds the step took to execute."""

    inputs: Optional[List[NamedValue]] = None
    """Inputs and their values passed to the test."""

    outputs: Optional[List[NamedValue]] = None
    """Outputs and their values logged by the test."""

    keywords: Optional[List[str]] = None
    """Words or phrases associated with the step."""

    properties: Optional[Dict[str, str]] = None
    """Test step properties."""


class CreateStepRequest(BaseStepRequest):
    name: str
    """Step name."""

    children: Optional[List["CreateStepRequest"]] = None
    """Nested child steps."""



from datetime import datetime
from typing import Dict, List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel
from nisystemlink.clients.testmonitor.models._named_value import NamedValue
from nisystemlink.clients.testmonitor.models._status import Status
from nisystemlink.clients.testmonitor.models._step_data import StepData


class Step(JsonModel):
    name: Optional[str] = None
    """The name of the step."""

    step_type: Optional[str] = None
    """The type of the step."""

    step_id: Optional[str] = None
    """The ID of the step."""

    parent_id: Optional[str] = None
    """The ID of the parent step."""

    result_id: Optional[str] = None
    """The ID of the result associated with the step."""

    path: Optional[str] = None
    """The path of the step."""

    path_ids: Optional[List[str]] = None
    """The IDs of the steps in the path."""

    status: Optional[Status] = None
    """The status of the step."""

    total_time_in_seconds: Optional[float] = None
    """The total time in seconds the step took to execute."""

    started_at: Optional[datetime] = None
    """The ISO-8601 formatted timestamp indicating when the step started."""

    updated_at: Optional[datetime] = None
    """The ISO-8601 formatted timestamp indicating when the step was last updated."""

    inputs: Optional[List[NamedValue]] = None
    """Inputs and their values passed to the test."""

    outputs: Optional[List[NamedValue]] = None
    """Outputs and their values logged by the test."""

    data_model: Optional[str] = None
    """Custom string defining the model of the data object."""

    data: Optional[StepData] = None
    """Data returned by the test step."""

    has_children: Optional[bool] = None
    """Indicates if the step has child steps."""

    workspace: Optional[str] = None
    """The workspace the test step belongs to."""

    keywords: Optional[List[str]] = None
    """Words or phrases associated with the test step."""

    properties: Optional[Dict[str, str]] = None
    """Test step properties."""



from typing import List, Optional

from nisystemlink.clients.core._api_error import ApiError
from nisystemlink.clients.core._uplink._json_model import JsonModel

from ._asset import Asset
from ._create_asset_request import CreateAssetRequest


class CreateAssetsPartialSuccessResponse(JsonModel):
    """Model for create Assets Partial Success Response."""

    assets: Optional[List[Asset]] = None
    """Gets or sets array of created assets."""

    failed: Optional[List[CreateAssetRequest]] = None
    """Gets or sets array of assets create requests that failed."""

    error: Optional[ApiError] = None
    """Gets or sets error if the create operation failed."""



from datetime import datetime
from typing import Dict, List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel
from pydantic import Field

from ._asset_calibration import (
    CalibrationStatus,
    ExternalCalibration,
    SelfCalibration,
    TemperatureSensor,
)
from ._asset_location import AssetLocation
from ._asset_types import AssetBusType, AssetDiscoveryType, AssetType


class Asset(JsonModel):
    """Model for an object describing an asset with all of its properties."""

    model_name: Optional[str] = None
    """Gets or sets model name of the asset."""

    model_number: Optional[int] = None
    """Gets or sets model number of the asset."""

    serial_number: Optional[str] = None
    """Gets or sets serial number of the asset."""

    vendor_name: Optional[str] = None
    """Gets or sets vendor name of the asset."""

    vendor_number: Optional[int] = None
    """Gets or sets vendor number of the asset."""

    bus_type: Optional[AssetBusType] = None
    """Gets or sets all supported bus types for an asset."""

    name: Optional[str] = None
    """Gets or sets name of the asset."""

    asset_type: Optional[AssetType] = None
    """Gets or sets all supported asset types."""

    discovery_type: Optional[AssetDiscoveryType] = None
    """Gets or sets the discovery type."""

    firmware_version: Optional[str] = None
    """Gets or sets firmware version of the asset."""

    hardware_version: Optional[str] = None
    """Gets or sets hardware version of the asset."""

    visa_resource_name: Optional[str] = None
    """Gets or sets VISA resource name of the asset."""

    temperature_sensors: Optional[List[TemperatureSensor]] = None
    """Gets or sets an array of temperature sensor information."""

    supports_self_calibration: Optional[bool] = None
    """Gets or sets whether the asset supports self-calibration."""

    supports_external_calibration: Optional[bool] = None
    """Gets or sets whether the asset supports external calibration."""

    custom_calibration_interval: Optional[int] = None
    """Gets or sets the interval represented in months used for computing calibration due date."""

    self_calibration: Optional[SelfCalibration] = None
    """Gets or sets the last self-calibration of the asset."""

    is_NI_asset: Optional[bool] = Field(alias="isNIAsset", default=None)
    """Gets or sets whether this asset is an NI asset (true) or a third-party asset (false)."""

    id: Optional[str] = None
    """Gets or sets unique identifier of the asset."""

    location: Optional[AssetLocation] = None
    """Model for information about the asset location, presence and the connection status of the system"""

    calibration_status: Optional[CalibrationStatus] = None
    """Gets or sets the calibration category the asset belongs to based on the next due calibration date."""

    is_system_controller: Optional[bool] = None
    """Gets or sets whether this asset represents a System Controller."""

    external_calibration: Optional[ExternalCalibration] = None
    """Gets or sets the last external calibration of the asset."""

    workspace: Optional[str] = None
    """Gets or sets the ID of the workspace."""

    properties: Optional[Dict[str, str]] = None
    """	Gets or sets key-value-pair metadata associated with an asset."""

    keywords: Optional[List[str]] = None
    """Gets or sets words or phrases associated with an asset."""

    last_updated_timestamp: Optional[datetime] = None
    """Gets or sets ISO-8601 formatted timestamp specifying the last date that the asset has had a property update."""

    file_ids: Optional[List[str]] = None
    """Gets or sets all files linked to the asset."""

    supports_self_test: Optional[bool] = None
    """Gets or sets whether the asset supports self-test."""

    supports_reset: Optional[bool] = None
    """Gets or sets whether the asset supports reset."""

    part_number: Optional[str] = None
    """Gets or sets part number of the asset."""

    out_for_calibration: Optional[bool] = None
    """Get or set whether the asset is out for calibration."""



from enum import Enum
from typing import List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel
from nisystemlink.clients.core._uplink._with_paging import WithPaging
from nisystemlink.clients.spec.models._specification import Specification


class SpecificationProjection(str, Enum):
    """The allowed projections for query.

    When using projection, only the fields specified by the projection element will be included in
    the response.
    """

    ID = "ID"
    PRODUCT_ID = "PRODUCT_ID"
    SPEC_ID = "SPEC_ID"
    NAME = "NAME"
    CATEGORY = "CATEGORY"
    TYPE = "TYPE"
    SYMBOL = "SYMBOL"
    BLOCK = "BLOCK"
    LIMIT = "LIMIT"
    UNIT = "UNIT"
    CONDITION_NAME = "CONDITION_NAME"
    CONDITION_VALUES = "CONDITION_VALUES"
    CONDITION_UNIT = "CONDITION_UNIT"
    CONDITION_TYPE = "CONDITION_TYPE"
    KEYWORDS = "KEYWORDS"
    PROPERTIES = "PROPERTIES"
    WORKSPACE = "WORKSPACE"
    CREATED_AT = "CREATED_AT"
    CREATED_BY = "CREATED_BY"


class SpecificationOrderBy(Enum):
    """The valid ways to order the response to a spec query."""

    ID = "ID"
    SPEC_ID = "SPEC_ID"


class QuerySpecificationsRequest(JsonModel):
    """The request to query specifications."""

    product_ids: List[str]
    """IDs of the products to query the specifications for."""

    take: Optional[int] = None
    """Maximum number of specifications to return in the current API response.

    Uses the default if the specified value is negative. The default value is `1000` specs.
    """

    continuation_token: Optional[str] = None
    """Allows users to continue the query at the next specification that matches the given criteria.

    To retrieve the next batch of specifications, pass the continuation token from the previous
    batch in the next request. The service responds with the next batch of data and provides a new
    continuation token. To paginate results, continue sending requests with the newest continuation
    token provided in each response.
    """

    filter: Optional[str] = None
    """
    The specification query filter in Dynamic Linq format.

    Allowed properties in the filter are:
    - `specId`: String representing the SpecID of a specification.
    - `name`: String representing the name of a specification.
    - `category`: String representing the category of a specification.
    - `type`: String enumeration representing the type of the specification.
        Possible values are : PARAMETRIC, FUNCTIONAL
    - `block`: String representing the block of a specification.
    - `symbol`: String representing the symbol of a specification.
    - `unit`: String representing the unit of a specification.
    - `workspace`: String representing the ID of the workspace the specification belongs to.
    - `createdBy`: String representing the ID of the user who created the specification.
    - `createdAt`: ISO-8601 formatted UTC timestamp indicating when the specification was created.
    - `updatedBy`: String representing the ID of the user who updated the specification.
    - `updatedAt`: ISO-8601 formatted UTC timestamp indicating when the specification was updated.

    See [Dynamic Linq](https://github.com/ni/systemlink-OpenAPI-documents/wiki/Dynamic-Linq-Query-Language)
    documentation for more details.
    """

    projection: Optional[List[SpecificationProjection]] = None
    """Specifies the fields to include in the returned specifications.

    Fields you do not specify are excluded. Returns all fields if no value is specified.
    """

    order_by: Optional[SpecificationOrderBy] = None
    """Specifies the field to use to sort specifications.

    By default, specifications are sorted by `ID`.
    """
    order_by_descending: Optional[bool] = None
    """Specifies whether to return the specifications in descending order.

    By default, this value is `false` and specifications are sorted in ascending order.
    """


class PagedSpecifications(WithPaging):
    """The list of matching specifications and a continuation token to get the next items."""

    specs: Optional[List[Specification]] = None
    """List of queried specifications.

    An empty list indicates that there are no specifications meeting the criteria provided in the
    request.
    """



"""Utilities for getting feed details."""

from typing import List, Optional

from nisystemlink.clients.feeds.models import Feed


def get_feed_by_name(
    feeds: List[Feed],
    name: str,
) -> Optional[Feed]:
    """Get feed information from the list of feeds using `name`.

    Args:
        feeds (List[Feed]): List of feeds.
        name (str): Feed name.

    Returns:
        Optional[Feed]: Feed information.
    """
    for feed in feeds:
        if feed.name == name and feed.id:
            return feed
    return None



from enum import Enum
from typing import List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel
from nisystemlink.clients.core._uplink._with_paging import WithPaging


class StepOrderBy(str, Enum):
    """The valid ways to order steps query response.

    This contains only limited fields available in a Step.
    """

    NAME = "NAME"
    STEP_TYPE = "STEP_TYPE"
    STEP_ID = "STEP_ID"
    PARENT_ID = "PARENT_ID"
    RESULT_ID = "RESULT_ID"
    PATH = "PATH"
    TOTAL_TIME_IN_SECONDS = "TOTAL_TIME_IN_SECONDS"
    STARTED_AT = "STARTED_AT"
    UPDATED_AT = "UPDATED_AT"
    DATA_MODEL = "DATA_MODEL"


class StepField(str, Enum):
    """The valid field values that can be queried.

    This contains only limited fields available in a Step.
    """

    NAME = "NAME"
    STEP_TYPE = "STEP_TYPE"
    STEP_ID = "STEP_ID"
    PARENT_ID = "PARENT_ID"
    RESULT_ID = "RESULT_ID"
    PATH = "PATH"
    TOTAL_TIME_IN_SECONDS = "TOTAL_TIME_IN_SECONDS"
    STARTED_AT = "STARTED_AT"
    UPDATED_AT = "UPDATED_AT"
    DATA_MODEL = "DATA_MODEL"


class StepProjection(str, Enum):
    """An enumeration of all fields in a Step.

    This enumeration is used to specify the fields to project in a step query.
    """

    NAME = "NAME"
    STEP_TYPE = "STEP_TYPE"
    STEP_ID = "STEP_ID"
    PARENT_ID = "PARENT_ID"
    RESULT_ID = "RESULT_ID"
    PATH = "PATH"
    PATH_IDS = "PATH_IDS"
    STATUS = "STATUS"
    TOTAL_TIME_IN_SECONDS = "TOTAL_TIME_IN_SECONDS"
    STARTED_AT = "STARTED_AT"
    UPDATED_AT = "UPDATED_AT"
    INPUTS = "INPUTS"
    OUTPUTS = "OUTPUTS"
    DATA_MODEL = "DATA_MODEL"
    DATA = "DATA"
    HAS_CHILDREN = "HAS_CHILDREN"
    WORKSPACE = "WORKSPACE"
    KEYWORDS = "KEYWORDS"
    PROPERTIES = "PROPERTIES"


class QueryStepsBase(JsonModel):
    filter: Optional[str] = None
    """The step query filter in Dynamic Linq format."""

    substitutions: Optional[List[str]] = None
    """String substitutions into the `filter`.

    Makes substitutions in the query filter expression. Substitutions for the query expression are
    indicated by non-negative integers that are prefixed with the "at" symbol. Each substitution in
    the given expression will be replaced by the element at the corresponding index (zero-based) in
    this list. For example, "@0" in the filter expression will be replaced with the element at the
    zeroth index of the substitutions list.
    """


class QueryStepsRequest(QueryStepsBase, WithPaging):
    order_by: Optional[StepOrderBy] = None
    """Specifies the fields to use to sort the steps."""

    descending: Optional[bool] = None
    """Specifies whether to return the steps in descending order."""

    take: Optional[int] = None
    """Maximum number of steps to return in the current API response."""

    return_count: Optional[bool] = None
    """If true, the response will include a count of all steps matching the filter.

    By default, this value is `False` and count is not returned. Note that returning the count may
    incur performance penalties as the service may have to do a complete walk of the database to
    compute count.
    """

    result_filter: Optional[str] = None
    """The result query filter in Dynamic Linq format."""

    result_substitutions: Optional[List[str]] = None
    """String substitutions into the `result_filter`."""

    projection: Optional[List[StepProjection]] = None
    """Specifies the step fields to project. When a field value is given here,
    the corresponding field will be present in all returned steps, and all
    unspecified fields will be excluded. If no projection is specified,
    all step fields will be returned.
    """


class QueryStepValuesRequest(QueryStepsBase):
    field: StepField
    """The step field to return for this query."""

    starts_with: Optional[str] = None
    """Only return string parameters prefixed by this value (case sensitive)."""



from enum import Enum


class FileQueryOrderBy(Enum):
    """Order Files Query Response by Metadata."""

    CREATED = "created"
    ID = "id"
    SIZE = "size"
    LAST_UPDATED_TIMESTAMP = "lastUpdatedTimestamp"



from ._create_test_plan_request import CreateTestPlanRequest
from ._create_test_plans_partial_success_response import (
    CreateTestPlansPartialSuccessResponse,
)
from ._dashboard import Dashboard, DashboardUrl
from ._update_test_plan_request import UpdateTestPlanRequest
from ._update_test_plans_request import UpdateTestPlansRequest
from ._update_test_plans_response import UpdateTestPlansResponse
from ._query_test_plans_request import QueryTestPlansRequest, TestPlanField
from ._paged_test_plans import PagedTestPlans
from ._execution_event import (
    ExecutionEvent,
    NotebookExecutionEvent,
    JobExecutionEvent,
    ManualExecutionEvent,
)
from ._order_by import OrderBy
from ._test_plan import TestPlan
from ._state import State
from ._execution_definition import (
    ExecutionDefinition,
    ManualExecution,
    JobExecution,
    NoneExecution,
    NotebookExecution,
    Job,
)
from ._schedule_test_plans_request import ScheduleTestPlansRequest
from ._schedule_test_plan_request import ScheduleTestPlanRequest
from ._schedule_test_plans_response import ScheduleTestPlansResponse

from ._test_plan_templates import TestPlanTemplateBase, TestPlanTemplate
from ._create_test_plan_templates_partial_success_response import (
    CreateTestPlanTemplatePartialSuccessResponse,
)
from ._delete_test_plan_templates_partial_success_response import (
    DeleteTestPlanTemplatesPartialSuccessResponse,
)
from ._query_test_plan_templates_request import (
    QueryTestPlanTemplatesRequest,
    TestPlanTemplateField,
    TestPlanTemplateOrderBy,
)
from ._paged_test_plan_templates import PagedTestPlanTemplates
from ._create_test_plan_template_request import CreateTestPlanTemplateRequest

# flake8: noqa



from typing import List, Optional

from nisystemlink.clients.core._uplink._with_paging import WithPaging

from ._column_order_by import ColumnOrderBy
from ._query_table_data_base import QueryTableDataBase


class QueryTableDataRequest(QueryTableDataBase, WithPaging):
    """Contains the filtering and sorting options to use when querying table data."""

    order_by: Optional[List[ColumnOrderBy]] = None
    """A list of columns to order the results by. Multiple columns may be
    specified to order rows that have the same value for prior columns. The
    columns used for sorting do not need to be included in the columns list, in
    which case they are not returned. If ``order_by`` is not specified, then the
    order in which results are returned is undefined."""

    take: Optional[int] = None
    """Limits the returned list to the specified number of results."""



from typing import Dict, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel
from pydantic import Field

from ._execution import (
    ExecutionPriority,
    ExecutionResourceProfile,
    ReportSettings,
)


class CreateExecutionRequest(JsonModel):
    """Creation information about an execution of a Jupyter notebook."""

    notebook_id: str = Field(..., min_length=1)
    """The ID of the notebook to execute."""

    parameters: Optional[Dict[str, Optional[str]]] = None
    """The input parameters for this execution of the notebook. The keys are strings and the values can be of any
    valid JSON type."""

    workspace_id: str = Field(..., min_length=1)
    """The ID of the workspace this execution belongs to."""

    timeout: Optional[int] = None
    """The number of seconds the execution runs before it aborts if uncompleted. The timer starts once status
    is IN_PROGRESS. 0 means infinite."""

    result_cache_period: Optional[int] = None
    """The period of time, in seconds, when the result of a previous notebook execution can be used as the
    result of the current notebook execution. Results will only be reused if the notebook and input parameters match.
    A value of 0 means do not reuse results from any previous executions. A value of -1 means always reuse results if
    possible and return an error if not possible. This prevents actual execution of a notebook."""

    report_settings: Optional[ReportSettings] = None
    """Settings of the Report"""

    client_requests_id: Optional[str] = None
    """The client request ID unique for each execution to be created. This is provided by the caller to be able
    to track executions that could not be created. The error response will contain this in the resourceId field.
    If not provided, the notebookId is used as the resourceId."""

    priority: Optional[ExecutionPriority] = None
    """Execution priority. Can be one of Low, Medium or High."""

    resource_profile: Optional[ExecutionResourceProfile] = None
    """Resource profile of the execution."""



from __future__ import annotations

from datetime import datetime
from typing import Dict, List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel


class PackageMetadata(JsonModel):
    """Package Meta data."""

    package_name: Optional[str] = None
    """The name of the package."""
    version: Optional[str] = None
    """The version number of the package."""
    architecture: Optional[str] = None
    """The architecture of the package."""
    breaks: Optional[List[str]] = None
    """Information about other packages this package breaks."""
    conflicts: Optional[List[str]] = None
    """Information about other packages this package conflicts with."""
    depends: Optional[List[str]] = None
    """Information about other packages this package depends on."""
    description: Optional[str] = None
    """The description of the package."""
    enhances: Optional[List[str]] = None
    """Information about other packages this package enchances."""
    essential: Optional[bool] = None
    """True if the package is essential."""
    file_name: Optional[str] = None
    """The file name of the package. Depending on the selected platform,
    the following package extensions are available:
    .nipkg for windows feeds, .ipk and .deb for ni-linux-rt feeds."""
    homepage: Optional[str] = None
    """The website of the maintainers for the package."""
    installed_size: Optional[int] = None
    """The size of the package after install."""
    maintainer: Optional[str] = None
    """The maintainer of the package (name and email address)."""
    predepends: Optional[List[str]] = None
    """Information about other packages this package predepends."""
    priority: Optional[int] = None
    """The priority of the package."""
    provides: Optional[List[str]] = None
    """Information about other packages that this package provides."""
    recommends: Optional[List[str]] = None
    """Information about other packages this package recommends."""
    release_notes: Optional[str] = None
    """The release notes of the package."""
    replaces: Optional[List[str]] = None
    """Information about other packages this package replaces."""
    section: Optional[str] = None
    """The application area of the package."""
    size: Optional[int] = None
    """The size (in bytes) of the package."""
    source: Optional[str] = None
    """The source of the package."""
    suggests: Optional[List[str]] = None
    """Information about other packages this package suggests."""
    tags: Optional[str] = None
    """The tags of the package."""
    attributes: Optional[Dict[str, str]] = None
    """The attributes of the package."""


class Package(JsonModel):
    """Package model."""

    id: Optional[str] = None
    """Gets or sets the ID of this package. This is used to reference this package in the service."""
    file_name: Optional[str] = None
    """The name of the file in this package."""
    feed_id: Optional[str] = None
    """The ID of the feed this package is associated with."""
    workspace: Optional[str] = None
    """The ID of the workspace this package belongs to.
    The workspace of a package is the workspace of feed this package is associated with."""
    updated_at: Optional[datetime] = None
    """The date of the latest package update."""
    created_at: Optional[datetime] = None
    """The date when the package was created at."""
    metadata: Optional[PackageMetadata] = None
    """Package meta data."""



from typing import List, Optional

from nisystemlink.clients.core import ApiError
from nisystemlink.clients.core._uplink._json_model import JsonModel


class DeleteResultsPartialSuccess(JsonModel):
    """The result of deleting multiple results when one or more results could not be deleted."""

    ids: List[str]
    """The IDs of the results that were successfully deleted."""

    failed: Optional[List[str]] = None
    """The IDs of the results that could not be deleted."""

    error: Optional[ApiError] = None
    """The error that occurred when deleting the results."""



from enum import Enum
from typing import List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel

from ._query_table_data_base import QueryTableDataBase


class DecimationMethod(str, Enum):
    """Represents the different methods that can be used to decimate data."""

    Lossy = "LOSSY"
    """Creates an ``x_column`` ordered set and returns an uniformly distributed
    sample of rows with as many rows as the number specified as ``intervals.``"""

    MaxMin = "MAX_MIN"
    """Creates an ``x_column`` ordered set which will be divided in the number of
    ``intervals`` specified. For each of the intervals, the maximum and minimum
    values for all the columns specified in ``y_columns`` will be returned."""

    EntryExit = "ENTRY_EXIT"
    """Creates an ``x_column`` ordered set which will be divided in the number of
    ``intervals`` specified. For each of the intervals, the first and last row
    within the interval will be returned in addition to the maximum and minimum
    values for all the columns specified in ``y_columns``."""


class DecimationOptions(JsonModel):
    """Contains the parameters to use for data decimation."""

    x_column: Optional[str] = None
    """The name of the column that will be used as the x-axis for decimating the
    data. The column in the table that was specified as Index will be used if
    this field is excluded. Only numeric columns are supported. i.e. ``INT32``,
    ``INT64``, ``FLOAT32``, ``FLOAT64`` and ``TIMESTAMP``."""

    y_columns: Optional[List[str]] = None
    """A list of columns to decimate by. This property is only needed when the
    specified method is ``MAX_MIN`` or ``ENTRY_EXIT``. Only numeric columns are
    supported. i.e. ``INT32``, ``INT64``, ``FLOAT32``, ``FLOAT64`` and
    ``TIMESTAMP``."""

    intervals: Optional[int] = None
    """Number of intervals to use for decimation. Defaults to 1000."""

    method: Optional[DecimationMethod] = None
    """Specifies the method used to decimate the data. Defaults to
    :class:`DecimationMethod.Lossy`"""


class QueryDecimatedDataRequest(QueryTableDataBase):
    """Specifies the columns, filters and decimation parameters for a query."""

    decimation: Optional[DecimationOptions] = None
    """The decimation options to use when querying data. If not specified, the
    default is to use :class:`DecimationMethod.Lossy` with 1000 intervals over
    the table's index column."""



from enum import Enum


class AssetBusType(Enum):
    """All supported bus types for an asset."""

    BUILT_IN_SYSTEM = "BUILT_IN_SYSTEM"
    PCI_PXI = "PCI_PXI"
    USB = "USB"
    GPIB = "GPIB"
    VXI = "VXI"
    SERIAL = "SERIAL"
    TCP_IP = "TCP_IP"
    CRIO = "CRIO"
    SCXI = "SCXI"
    CDAQ = "CDAQ"
    SWITCH_BLOCK = "SWITCH_BLOCK"
    SCC = "SCC"
    FIRE_WIRE = "FIRE_WIRE"
    ACCESSORY = "ACCESSORY"
    CAN = "CAN"
    SWITCH_BLOCK_DEVICE = "SWITCH_BLOCK_DEVICE"
    SLSC = "SLSC"


class AssetType(Enum):
    """All supported asset types."""

    GENERIC = "GENERIC"
    DEVICE_UNDER_TEST = "DEVICE_UNDER_TEST"
    FIXTURE = "FIXTURE"
    SYSTEM = "SYSTEM"


class AssetDiscoveryType(Enum):
    """All discovery types."""

    MANUAL = "MANUAL"
    AUTOMATIC = "AUTOMATIC"



from typing import List, Optional

from nisystemlink.clients.core._api_error import ApiError
from nisystemlink.clients.core._uplink._json_model import JsonModel


class DeleteTestPlanTemplatesPartialSuccessResponse(JsonModel):
    """The result of deleting multiple test plan templates
    when one or more test plan templates could not be deleted.
    """

    deleted_test_plan_template_ids: List[str]
    """The IDs of the test plan template that could not be deleted."""

    failed_test_plan_template_ids: Optional[List[str]] = None
    """The IDs of the test plan template that could not be deleted."""

    error: Optional[ApiError] = None
    """The error that occurred when deleting the test plan template."""



from typing import List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel

from ._update_test_plan_request import UpdateTestPlanRequest


class UpdateTestPlansRequest(JsonModel):
    """Represents the request body for updating multiple test plans."""

    test_plans: List[UpdateTestPlanRequest]
    """A list of test plans to update."""

    replace: Optional[bool] = None
    """Whether to replace the existing test plans or update them."""



from typing import Any, Callable, Dict, List, Optional

import pandas as pd
from nisystemlink.clients.testmonitor.models import (
    Measurement,
    Result,
    Step,
    StepProjection,
)
from nisystemlink.clients.testmonitor.utilities.constants import DataFrameHeaders


def has_name_and_measurement(measurement: Measurement) -> bool:
    """Checks if a step data parameter is measurement data by ensuring it has both
    'name' and 'measurement' fields.

    Args:
        measurement: A measurement data object

    Returns:
        bool: True if the measurement has both 'name' and 'measurement' fields, False otherwise.
    """
    return measurement.name is not None and measurement.measurement is not None


def convert_results_to_dataframe(
    results: List[Result], set_id_as_index: bool = True
) -> pd.DataFrame:
    """Creates a Pandas DataFrame for the results.

    Args:
        results: The list of results to be included in the dataframe.
        set_id_as_index: If true (default value), result id will be set as index for the dataframe.
                         If false, index will not be set.

    Returns:
        A Pandas DataFrame with the each result fields having a separate column.
        Following fields are split into sub-columns.
            - status_type_summary: All the entries will be split into separate columns.
            For example, status_type_summary.LOOPING, status_type_summary.PASSED, etc
            - Properties: All the properties will be split into separate columns. For example,
            properties.property1, properties.property2, etc.
    """
    results_dict = []
    for result in results:
        data = result.model_dump(exclude_none=True)
        __normalize_status(data)
        results_dict.append(data)

    normalized_dataframe = pd.json_normalize(results_dict, sep=".")
    normalized_dataframe = __format_results_columns(
        results_dataframe=normalized_dataframe
    )
    if set_id_as_index and "id" in normalized_dataframe.columns:
        normalized_dataframe.set_index("id", inplace=True)

    return normalized_dataframe


def convert_steps_to_dataframe(
    steps: List[Step],
    is_valid_measurement: Optional[
        Callable[[Measurement], bool]
    ] = has_name_and_measurement,
) -> pd.DataFrame:
    """Converts a list of steps into a normalized dataframe.

    Args:
        steps: A list of steps.
        is_valid_measurement: Optional callback function that checks if a step data parameter is a
            valid measurement so that only those are included in the returned dataframe. The method takes
            a measurement as input and returns a boolean value.
            The default behavior is to consider only measurement data that have both 'name' and 'measurement'
            fields with values as valid measurements.
            If none of the measurements have the required fields, the step data parameters will not
            appear in the dataframe.
            If the callback function is set to None, all step data parameters will be included in the dataframe.

    Returns:
        DataFrame:
            - A Pandas DataFrame containing the steps data. The DataFrame would consist of all the
            fields in the input steps.
            - A new column would be created for unique `properties` across all steps. The property
            columns would be named in the format `properties.property_name`.
            - A new column would be created for unique `Inputs` and `Outputs` across all steps. The columns
            would be named in the format `inputs.input_name` and `outputs.output_name` respectively.
            - The column headers for the step data parameters would differ based on the callback function. If
            the None is passed for the callback function, the column would be prefixed with `data.parameters.`.
            If the callback function is set, the column would be prefixed with `data.measurement.`.
    """
    DATA_PARAMETERS_PREFIX = (
        "data.parameters" if is_valid_measurement is None else "data.measurement"
    )
    step_dicts = __convert_steps_to_dict(steps, is_valid_measurement)
    steps_dataframe = pd.json_normalize(step_dicts, sep=".")
    steps_dataframe = __explode_and_normalize(
        steps_dataframe, "data.parameters", f"{DATA_PARAMETERS_PREFIX}."
    )
    grouped_columns = __group_step_columns(steps_dataframe.columns)
    return steps_dataframe.reindex(columns=grouped_columns, copy=False)


def __normalize_status(
    data: Dict[str, Any],
) -> None:
    """Normalizes the status object into a string.

    Args:
        data: Dictionary containing status information.

    """
    status = data.get("status", {})
    if status:
        if status.get("status_type") == "CUSTOM":
            data["status"] = status.get("status_name", None)
        else:
            data["status"] = getattr(status.get("status_type", None), "value", None)


def __format_results_columns(results_dataframe: pd.DataFrame) -> pd.DataFrame:
    """Format results column to keep properties at the end.

    Args:
        results_dataframe: Dataframe of results.

    Returns:
        Formatted dataframe of results.
    """
    column_headers = results_dataframe.columns.to_list()
    standard_column_headers = [
        header for header in column_headers if __is_standard_column_header(header)
    ]
    status_type_summary_header = [
        header
        for header in column_headers
        if __is_status_type_summary_header(header=header)
    ]
    properties_headers = [
        header for header in column_headers if __is_property_header(header=header)
    ]
    standard_column_headers += status_type_summary_header + properties_headers

    return results_dataframe.reindex(columns=standard_column_headers, copy=False)


def __is_standard_column_header(header: str) -> bool:
    """Check if column header is not status type summary or property.

    Args:
        header: column header for results dataframe.

    Returns:
        True if header doesn't start with 'status_type_summary.', 'properties.'. Else returns false.

    """
    return not (
        __is_status_type_summary_header(header=header)
        or __is_property_header(header=header)
    )


def __is_status_type_summary_header(header: str) -> bool:
    """Check if column header is not a status type summary.

    Args:
        header: column header for results dataframe.

    Returns:
        True if header contains 'status_type_summary.'. Else returns false.

    """
    return header.startswith(DataFrameHeaders.STATUS_TYPE_SUMMARY_HEADER_PREFIX)


def __is_property_header(header: str) -> bool:
    """Check if column header is not a property.

    Args:
        header: column header for results dataframe.

    Returns:
        True if header contains 'properties.'. Else returns false.

    """
    return header.startswith(DataFrameHeaders.PROPERTY_COLUMN_HEADER_PREFIX)


def __convert_steps_to_dict(
    steps: List[Step],
    is_valid_measurement: Optional[Callable[[Measurement], bool]],
) -> List[Dict[str, Any]]:
    """Converts a list of steps to dictionaries, excluding None values.

    Args:
        steps: A list of steps.
        is_valid_measurement: Optional callback function that checks if a step data
            parameter is a valid measurement so that only those are included in the returned dataframe.

    Returns:
        List[Dict[str, Any]]: A list of dictionaries containing step information.
    """
    steps_dict = []
    for step in steps:

        single_step_dict = step.model_dump(exclude_none=True)
        __filter_invalid_measurements(single_step_dict, step, is_valid_measurement)
        __normalize_inputs_outputs(single_step_dict, step)
        __normalize_status(single_step_dict)
        steps_dict.append(single_step_dict)
    return steps_dict


def __filter_invalid_measurements(
    step_dict: Dict[str, Any],
    step: Step,
    is_valid_measurement: Optional[Callable[[Measurement], bool]],
) -> None:
    """Gets data parameters from the step dictionary and filters it based on the callback function.

    Args:
        step_dict: A dictionary with step information.
        step: A Step object containing data parameters.
        is_valid_measurement: Optional callback function to check if a measurement is valid. The method takes
            a Measurement as input and returns a boolean value. The default behavior is to consider only parameters
            that have both 'name' and 'measurement' fields with values as valid measurements.

    Returns:
        None: The function modifies step dictionary in place with filtered data parameters.
    """
    if step.data and step.data.parameters and is_valid_measurement is not None:
        valid_measurement_parameters = []
        for measurement in step.data.parameters:
            if measurement and is_valid_measurement(measurement):
                valid_measurement_parameters.append(measurement)

        step_dict["data"]["parameters"] = [
            measurement.model_dump(exclude_none=True)
            for measurement in valid_measurement_parameters
        ]


def __normalize_inputs_outputs(
    step_dict: Dict[str, Any],
    step: Step,
) -> None:
    """Normalizes the input and output fields by converting them into dictionaries.

    Args:
        step_dict: A dictionary with step information.
        step: A Step object containing inputs and outputs.

    Returns:
        None: The function modifies step_dict in place with normalized inputs and outputs.
    """
    STEP_INPUTS = StepProjection.INPUTS.lower()
    STEP_OUTPUTS = StepProjection.OUTPUTS.lower()
    if STEP_INPUTS in step_dict:
        step_dict[STEP_INPUTS] = (
            {item.name: item.value for item in step.inputs} if step.inputs else {}
        )
    if STEP_OUTPUTS in step_dict:
        step_dict[STEP_OUTPUTS] = (
            {item.name: item.value for item in step.outputs} if step.outputs else {}
        )


def __explode_and_normalize(
    dataframe: pd.DataFrame, column: str, prefix: str
) -> pd.DataFrame:
    """Explodes a specified column in the dataframe and normalizes its nested data.
    This function handles the process of exploding a column that contains lists or arrays,
    transforming each list element into a separate row. After exploding, it normalizes the
    nested data into flat columns using the specified prefix, making it easier to analyze
    and manipulate. The new columns are added to the original dataframe.

    Args:
        dataframe: The input DataFrame that contains the column to explode and normalize.
        column: The name of the column in the DataFrame that contains the list-like data to explode.
        prefix: The prefix to add to the new column names created during the normalization process.

    Returns:
        DataFrame:
        - A new DataFrame with the exploded rows and the normalized columns, all combined
        with the original data in the dataframe.
        - If the column is not found in the dataframe, the original dataframe is returned unchanged.
    """
    if column in dataframe:
        exploded_dataframe = dataframe.explode(column, ignore_index=True)
        normalized_dataframe = pd.json_normalize(
            exploded_dataframe.pop(column)
        ).add_prefix(prefix)
        return pd.concat([exploded_dataframe, normalized_dataframe], axis=1, copy=False)
    return dataframe


def __group_step_columns(dataframe_columns: List[str]) -> List[str]:
    """Groups and orders dataframe columns into predefined categories to maintain a consistent structure.
    When normalizing steps into a dataframe, new input, output, or property fields may be added at the end,
    disrupting the expected column order. This function ensures columns are grouped properly.

    Args:
        dataframe_columns: The list of all columns from the normalized dataframe.

    Returns:
        List[str]: A list containing grouped and ordered columns.
    """
    GENERAL_CATEGORIES = "general"
    CATEGORY_KEYS = DataFrameHeaders.CATEGORY_COLUMN_HEADERS
    grouped_columns: Dict[str, List[str]] = {category: [] for category in CATEGORY_KEYS}
    for column in dataframe_columns:
        column_lower = column.lower()
        key = next(
            (
                category
                for category in CATEGORY_KEYS[1:]
                if column_lower.startswith(category)
                and column != StepProjection.DATA_MODEL.lower()
            ),
            GENERAL_CATEGORIES,
        )
        grouped_columns[key].append(column)
    return [
        column
        for category_key in CATEGORY_KEYS
        for column in grouped_columns[category_key]
    ]



from enum import Enum
from typing import Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel


class AssetPresenceStatus(Enum):
    """Status of an asset's presence in a system."""

    INITIALIZING = "INITIALIZING"
    UNKNOWN = "UNKNOWN"
    NOT_PRESENT = "NOT_PRESENT"
    PRESENT = "PRESENT"


class SystemConnection(Enum):
    """Whether or not the minion is connected to the server and has updated the server with its data.
    To maintain compatibility with previous versions of SystemLink, the values
    [APPROVED, UNSUPPORTED, ACTIVATED] are considered equivalent to DISCONNECTED and
    [CONNECTED_UPDATE_PENDING, CONNECTED_UPDATE_SUCCESSFUL, CONNECTED_UPDATE_FAILED] are equivalent to CONNECTED.
    """

    APPROVED = "APPROVED"
    DISCONNECTED = "DISCONNECTED"
    CONNECTED_UPDATE_PENDING = "CONNECTED_UPDATE_PENDING"
    CONNECTED = "CONNECTED"
    CONNECTED_UPDATE_FAILED = "CONNECTED_UPDATE_FAILED"
    UNSUPPORTED = "UNSUPPORTED"
    ACTIVATED = "ACTIVATED"
    CONNECTED_UPDATE_SUCCESSFUL = "CONNECTED_UPDATE_SUCCESSFUL"


class AssetPresenceWithSystemConnection(JsonModel):
    """Model for the presence of an asset and the connection of the system in which it resides."""

    asset_presence: AssetPresenceStatus
    """Gets or sets the status of an asset's presence in a system."""

    system_connection: Optional[SystemConnection] = None
    """Gets or sets whether or not the minion is connected to the server and has updated the server with its data."""


class AssetPresence(JsonModel):
    """Model for the presence of an asset."""

    asset_presence: AssetPresenceStatus
    """Gets or sets the status of an asset's presence in a system."""


class _AssetLocation(JsonModel):
    """local model for information about the asset location, presence and the connection status of the system."""

    minion_id: Optional[str] = None
    """Gets or sets identifier of the minion where the asset is located."""

    physical_location: Optional[str] = None
    """Gets or sets the physical location of the asset."""

    parent: Optional[str] = None
    """Gets or sets the parent of the asset."""

    resource_uri: Optional[str] = None
    """Gets or sets identifier of a resource."""

    slot_number: Optional[int] = None
    """Gets or sets the number of the slot in which the asset is located."""


class AssetLocation(_AssetLocation):
    """Model for information about the asset location, presence and the connection status of the system."""

    state: AssetPresenceWithSystemConnection
    """Presence of an asset and the connection of the system in which it resides."""


class AssetLocationForCreate(_AssetLocation):
    """Model for information about the asset presence status of the system, used while create"""

    state: AssetPresence
    """Model for the presence of an asset."""



from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel


class CalibrationStatus(Enum):
    """Calibration category the asset belongs to based on the next due calibration date."""

    OK = "OK"
    APPROACHING_RECOMMENDED_DUE_DATE = "APPROACHING_RECOMMENDED_DUE_DATE"
    PAST_RECOMMENDED_DUE_DATE = "PAST_RECOMMENDED_DUE_DATE"
    OUT_FOR_CALIBRATION = "OUT_FOR_CALIBRATION"


class TemperatureSensor(JsonModel):
    """Temperature sensor information."""

    name: Optional[str] = None
    """Gets or sets sensor name."""

    reading: float
    """Gets or sets sensor reading."""


class SelfCalibration(JsonModel):
    temperature_sensors: Optional[List[TemperatureSensor]] = None
    """Gets or sets an array of temperature sensor information."""

    is_limited: Optional[bool] = None
    """Gets or sets whether the last self-calibration of the asset was a limited calibration."""

    date: datetime
    """Gets or sets ISO-8601 formatted timestamp specifying the last date the asset was self-calibrated."""


class CalibrationMode(Enum):
    """Whether SystemLink automatically discovered the calibration data for an asset or if it was manually entered."""

    AUTOMATIC = "AUTOMATIC"
    MANUAL = "MANUAL"


class ExternalCalibration(JsonModel):
    temperature_sensors: Optional[List[TemperatureSensor]] = None
    """Gets or sets an array of temperature sensor information."""

    is_limited: Optional[bool] = None
    """Gets or sets whether the last external calibration of the asset was a limited calibration."""

    date: datetime
    """Gets or sets ISO-8601 formatted timestamp specifying the last date the asset was externally calibrated."""

    recommended_interval: int
    """Gets or sets the manufacturer's recommended calibration interval in months."""

    next_recommended_date: datetime
    """Gets or sets ISO-8601 formatted timestamp specifying the recommended date for the next external calibration."""

    next_custom_due_date: Optional[datetime] = None
    """Gets or sets ISO-8601 formatted timestamp specifying the date for the next external calibration."""

    resolved_due_date: Optional[datetime] = None
    """Gets ISO-8601 formatted timestamp specifying the resolved due date for external calibration."""

    comments: Optional[str] = None
    """Gets or sets calibration comments provided by an operator."""

    entry_type: Optional[CalibrationMode] = None
    """Gets or sets whether automatically discovered the calibration data for an asset or manually entered."""



from typing import Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel


class CreateVirtualSystemResponse(JsonModel):
    """Model for create virtual system response containing the minion ID of the system which is created."""

    minionId: Optional[str] = None
    """The minion ID of the created virtual system."""



# -*- coding: utf-8 -*-

"""Implementation of HttpConfiguration."""

import pathlib
import urllib.parse
from typing import Dict, Optional


class HttpConfiguration:
    """Represents the configuration for accessing a SystemLink service over HTTP."""

    DEFAULT_TIMEOUT_MILLISECONDS = 60000
    """The default value of :attr:`timeout_milliseconds` to use when making API calls."""

    _SYSTEM_LINK_API_KEY_HEADER = "x-ni-api-key"

    def __init__(
        self,
        server_uri: str,
        api_key: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        cert_path: Optional[pathlib.Path] = None,
        workspace: Optional[str] = None,
        verify: bool = True,
    ) -> None:
        """Initialize a configuration.

        If neither ``api_key`` nor ``username`` and ``password`` are set, the
        configuration will use anonymous access, and any API calls that require
        authorization will fail.

        Args:
            server_uri: The scheme, host, and port (if not default) of the web server
                hosting the SystemLink service to connect to. Additional Uri properties
                such as ``urllib.parse.urlparse().path`` and
                ``urllib.parse.urlparse().query`` are ignored.
            api_key: The API key to send with requests.
            username: The name of the user to use when authorization is required.
            password: The user's password to use when authorization is required.
            cert_path: Local path to an SSL certificate file.
            workspace: ID of workspace to use for client operations.
            verify: Verify the security certificate for connection

        Raises:
            ValueError: if ``server_uri`` is missing scheme or host information.
            ValueError: if ``username`` or ``password`` is set, but not both.
        """
        uri = urllib.parse.urlsplit(server_uri)
        if not uri.scheme:
            raise ValueError(
                "Scheme (e.g. http) not included in server_uri: '{}'".format(uri)
            )
        if not uri.hostname:
            raise ValueError(
                "Host (e.g. foo.com) not included in server_uri: '{}'".format(uri)
            )
        self._server_uri = urllib.parse.urlunsplit(uri[:2] + ("", "", ""))

        self._api_keys = None  # type: Optional[Dict[str, str]]
        if api_key:
            self._api_keys = {self._SYSTEM_LINK_API_KEY_HEADER: api_key}
        elif username or password:
            if not username or not password:
                raise ValueError("If username or password is set, both must be set")
            self._username = username
            self._password = password

        self._cert_path = cert_path

        self._user_agent = ""  # type: Optional[str]

        self._timeout_ms = self.DEFAULT_TIMEOUT_MILLISECONDS

        self._workspace = workspace

        self._verify = verify

    @property
    def verify(self) -> bool:
        """Verify the security certificate for connection."""
        return self._verify

    @verify.setter
    def verify(self, value: bool) -> None:
        self._verify = value

    @property
    def timeout_milliseconds(self) -> int:  # noqa: D401
        """The number of milliseconds before a request times out with an error.

        Changing the timeout will not affect APIs that have already read the
        configuration.
        """
        return self._timeout_ms

    @timeout_milliseconds.setter
    def timeout_milliseconds(self, value: int) -> None:
        self._timeout_ms = value

    @property
    def user_agent(self) -> Optional[str]:  # noqa: D401
        """The string to pass the web server as the product name or names making the
        request, or None to use a library-specific default.

        Changing the user-agent will not affect APIs that have already read the
        configuration.
        """
        return self._user_agent or None

    @user_agent.setter
    def user_agent(self, value: Optional[str]) -> None:
        self._user_agent = value

    @property
    def api_keys(self) -> Optional[Dict[str, str]]:  # noqa: D401
        """The available API keys to use for authorization, or None if none were provided."""
        return dict(self._api_keys) if self._api_keys else None

    @property
    def server_uri(self) -> str:  # noqa: D401
        """The ``urllib.parse.urlparse().scheme``, ``urllib.parse.urlparse().hostname``,
        and ``urllib.parse.urlparse().port`` of the web server hosting the SystemLink
        service to connect to.

        Additional Uri properties such as ``urllib.parse.urlparse().path`` and
        ``urllib.parse.urlparse().query`` are ignored.
        """
        return self._server_uri

    @property
    def username(self) -> Optional[str]:  # noqa: D401
        """The username to use for HTTP authentication, or None if none was provided."""
        return self._username

    @property
    def password(self) -> Optional[str]:  # noqa: D401
        """The password to use for HTTP authentication, or None if none was provided."""
        return self._password

    @property
    def cert_path(self) -> Optional[pathlib.Path]:
        """Local path to an SSL certificate file."""
        return self._cert_path

    @property
    def workspace(self) -> Optional[str]:  # noqa: D401
        """ID of workspace to use for Client operations."""
        return self._workspace



# -*- coding: utf-8 -*-

"""Implementation of CloudHttpConfiguration."""

from nisystemlink.clients import core


class CloudHttpConfiguration(core.HttpConfiguration):
    """An :class:`HttpConfiguration` specifically for use with SystemLink Cloud."""

    _CLOUD_URI = "https://api.systemlinkcloud.com"

    def __init__(self, api_key: str) -> None:
        """Initialize a configuration for SystemLink Cloud using API key-based authentication.

        Args:
            api_key: The API key to send with requests.

        Raises:
            ValueError: if ``api_key`` is empty.
        """
        super().__init__(self._CLOUD_URI, api_key=api_key)



from typing import List

from nisystemlink.clients.core._uplink._with_paging import WithPaging

from ._table_metadata import TableMetadata


class PagedTables(WithPaging):
    """The response for a table query containing the matched tables."""

    tables: List[TableMetadata]
    """The list of tables returned by the query."""



from enum import Enum
from typing import List, Optional, Union

from nisystemlink.clients.core._uplink._json_model import JsonModel
from pydantic import StrictFloat, StrictInt, StrictStr


class ConditionType(Enum):
    """Conditions are either numeric or string type."""

    NUMERIC = "NUMERIC"
    """Numeric condition."""

    STRING = "STRING"
    """String condition."""


class ConditionRange(JsonModel):
    """Specifies the range of values that the condition must cover."""

    min: Optional[float] = None
    """Minimum value of the condition range."""

    max: Optional[float] = None
    """Maximum value of the condition range."""

    step: Optional[float] = None
    """Step value of the condition range."""


class ConditionValueBase(JsonModel):
    """The base type for conditions that can be represented in several styles."""

    condition_type: Optional[ConditionType] = None
    """Type of the Condition."""


class NumericConditionValue(ConditionValueBase):
    """A numeric condition.

    Numeric conditions can contain a combination of ranges and discrete lists.
    """

    range: Optional[List[ConditionRange]] = None
    """List of condition range values."""

    # StrictFloat and StrictInt are used here because discrete is a common property for
    # NumericConditionValue and StringConditionValue.
    # If float/int is used, pydantic converts string of number into float/int by default
    # when deserializing a StringCondition JSON and misinterprets it as NumericConditionValue type.
    discrete: Optional[List[Union[StrictFloat, StrictInt]]] = None
    """List of condition discrete values."""

    unit: Optional[str] = None
    """Unit of the condition."""


class StringConditionValue(ConditionValueBase):
    """A string condition.

    String conditions may only contain discrete lists of values.
    """

    # StrictStr is used here because discrete is a common property for
    # NumericConditionValue and StringConditionValue.
    # If str is used, pydantic converts any datatype into string by default when deserializing a
    # NumericCondition JSON and misinterprets it as StringConditionValue type.
    discrete: Optional[List[StrictStr]] = None
    """List of condition discrete values."""


class Condition(JsonModel):
    """A single condition."""

    name: Optional[str] = None
    """Name of the condition."""

    # Ideal approach is to set the dtype here as ConditionValue and use pydantic discriminator to
    # deserialize/serialize the JSON into correct ConditionValue sub types. But Union of dtypes is
    # used here as the discriminator field could be none when projection is used in query API.
    value: Optional[Union[NumericConditionValue, StringConditionValue]] = None
    """Value of the condition."""



from typing import List, Optional

from nisystemlink.clients.core._api_error import ApiError
from nisystemlink.clients.core._uplink._json_model import JsonModel

from ._schedule_test_plan_request import ScheduleTestPlanRequest
from ._test_plan import TestPlan


class ScheduleTestPlansResponse(JsonModel):
    """Represents the response returned after attempting to schedule one or more test plans."""

    scheduled_test_plans: Optional[List[TestPlan]] = None
    """(Optional) List of successfully scheduled test plans."""

    failed_test_plans: Optional[List[ScheduleTestPlanRequest]] = None
    """(Optional) List of test plan requests that failed to schedule."""

    error: Optional[ApiError] = None
    """The error that occurred when scheduling the test plans."""



import enum
from typing import List, Optional

from nisystemlink.clients.core._uplink._with_paging import WithPaging

from ._order_by import OrderBy


class TestPlanField(enum.Enum):
    """Model for an object describing an test plan with all of its properties."""

    __test__ = False

    ID = "ID"
    TEMPLATE_ID = "TEMPLATE_ID"
    NAME = "NAME"
    STATE = "STATE"
    SUBSTATE = "SUBSTATE"
    DESCRIPTION = "DESCRIPTION"
    ASSIGNED_TO = "ASSIGNED_TO"
    WORK_ORDER_ID = "WORK_ORDER_ID"
    WORK_ORDER_NAME = "WORK_ORDER_NAME"
    PART_NUMBER = "PART_NUMBER"
    DUT_ID = "DUT_ID"
    DUT_SERIAL_NUMBER = "DUT_SERIAL_NUMBER"
    TEST_PROGRAM = "TEST_PROGRAM"
    WORKSPACE = "WORKSPACE"
    CREATED_BY = "CREATED_BY"
    UPDATED_BY = "UPDATED_BY"
    SYSTEM_ID = "SYSTEM_ID"
    FIXTURE_IDS = "FIXTURE_IDS"
    PLANNED_START_DATE_TIME = "PLANNED_START_DATE_TIME"
    ESTIMATED_END_DATE_TIME = "ESTIMATED_END_DATE_TIME"
    ESTIMATED_DURATION_IN_SECONDS = "ESTIMATED_DURATION_IN_SECONDS"
    SYSTEM_FILTER = "SYSTEM_FILTER"
    DUT_FILTER = "DUT_FILTER"
    CREATED_AT = "CREATED_AT"
    UPDATED_AT = "UPDATED_AT"
    PROPERTIES = "PROPERTIES"
    FILE_IDS_FROM_TEMPLATE = "FILE_IDS_FROM_TEMPLATE"
    DASHBOARD = "DASHBOARD"
    EXECUTION_ACTIONS = "EXECUTION_ACTIONS"
    EXECUTION_HISTORY = "EXECUTION_HISTORY"
    DASHBOARD_URL = "DASHBOARD_URL"
    WORKFLOW = "WORKFLOW"


class QueryTestPlansRequest(WithPaging):
    """Represents the request body for querying test plans.
    Allows filtering, sorting, and pagination of test plan results.
    """

    filter: Optional[str] = None
    """A string expression to filter the test plans returned by the query."""

    take: Optional[int] = None
    """The maximum number of test plans to return in the response."""

    order_by: Optional[OrderBy] = None
    """The field name to use for sorting the test plans."""

    descending: Optional[bool] = None
    """Whether to sort the test plans in descending order."""

    return_count: Optional[bool] = None
    """Whether to include the total count of matching test plans in the response."""

    projection: Optional[List[TestPlanField]] = None
    """
    Gets or sets the projection to be used when retrieving the assets. If not specified,
    all properties will be returned.
    """



from ._create_products_partial_success import CreateProductsPartialSuccess
from ._delete_products_partial_success import DeleteProductsPartialSuccess
from ._paged_products import PagedProducts
from ._query_products_request import (
    ProductField,
    ProductOrderBy,
    ProductProjection,
    QueryProductsRequest,
    QueryProductValuesRequest,
)
from ._product import Product
from ._product_request import CreateProductRequest, UpdateProductRequest

# flake8: noqa
